from __future__ import annotations
from .geom import *
import interpolation.deepsvg.svglib.geom as geom
import re
import torch
from typing import List, Union
from xml.dom import minidom
import math
import shapely.geometry
import numpy as np

from .geom import union_bbox
from .svg_command import SVGCommand, SVGCommandMove, SVGCommandClose, SVGCommandBezier, SVGCommandLine, SVGCommandArc


COMMANDS = "MmZzLlHhVvCcSsQqTtAa"
COMMAND_RE = re.compile(r"([MmZzLlHhVvCcSsQqTtAa])")
FLOAT_RE = re.compile(r"[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?")


empty_command = SVGCommandMove(Point(0.))


class Orientation:
    COUNTER_CLOCKWISE = 0
    CLOCKWISE = 1


class Filling:
    OUTLINE = 0
    FILL = 1
    ERASE = 2


class SVGPath:
    def __init__(self, path_commands: List[SVGCommand] = None, origin: Point = None, closed=False, filling=Filling.OUTLINE):
        self.origin = origin or Point(0.)
        # print(path_commands)
        self.path_commands = path_commands
        self.closed = closed

        self.filling = filling

    @property
    def start_command(self):
        return SVGCommandMove(self.origin, self.start_pos)

    @property
    def start_pos(self):
        return self.path_commands[0].start_pos

    @property
    def end_pos(self):
        return self.path_commands[-1].end_pos

    def to_group(self, *args, **kwargs):
        from .svg_primitive import SVGPathGroup
        return SVGPathGroup([self], *args, **kwargs)

    def set_filling(self, filling=True):
        self.filling = Filling.FILL if filling else Filling.ERASE
        return self

    def __len__(self):
        return 1 + len(self.path_commands)

    def __getitem__(self, idx):
        if idx == 0:
            return self.start_command
        return self.path_commands[idx-1]

    def all_commands(self, with_close=True):
        close_cmd = [SVGCommandClose(self.path_commands[-1].end_pos.copy(), self.start_pos.copy())] if self.closed and self.path_commands and with_close \
                    else ()
        return [self.start_command, *self.path_commands, *close_cmd]

    def copy(self):
        return SVGPath([path_command.copy() for path_command in self.path_commands], self.origin.copy(), self.closed, filling=self.filling)

    @staticmethod
    def _tokenize_path(path_str):
        cmd = None
        for x in COMMAND_RE.split(path_str):
            if x and x in COMMANDS:
                cmd = x
            elif cmd is not None:
                yield cmd, list(map(float, FLOAT_RE.findall(x)))

    @staticmethod
    def from_xml(x: minidom.Element):
        stroke = x.getAttribute('stroke')
        dasharray = x.getAttribute('dasharray')
        stroke_width = x.getAttribute('stroke-width')

        fill = not x.hasAttribute("fill") or not x.getAttribute("fill") == "none"
        print(x.hasAttribute("fill"),'~~~~svgPathfill')
        if fill == True:
            fill = x.getAttribute("fill")
        # fill = x.hasAttribute("fill") if not x.getAttribute("fill") == "none" else False

        filling = Filling.OUTLINE if not x.hasAttribute("filling") else int(x.getAttribute("filling"))

        s = x.getAttribute('d')
        return SVGPath.from_str(s, fill=fill, filling=filling)

    @staticmethod
    def from_str(s: str, fill=False, filling=Filling.OUTLINE, add_closing=False):
        path_commands = []
        pos = initial_pos = Point(0.)
        prev_command = None
        for cmd, args in SVGPath._tokenize_path(s):
            cmd_parsed, pos, initial_pos = SVGCommand.from_str(cmd, args, pos, initial_pos, prev_command)
            prev_command = cmd_parsed[-1]
            path_commands.extend(cmd_parsed)

        return SVGPath.from_commands(path_commands, fill=fill, filling=filling, add_closing=add_closing)

    @staticmethod
    def from_tensor(tensor: torch.Tensor, allow_empty=False):
        return SVGPath.from_commands([SVGCommand.from_tensor(row) for row in tensor], allow_empty=allow_empty)

    @staticmethod
    def from_commands(path_commands: List[SVGCommand], fill=False, filling=Filling.OUTLINE, add_closing=False, allow_empty=False):
        from .svg_primitive import SVGPathGroup

        if not path_commands:
            return SVGPathGroup([])

        svg_paths = []
        svg_path = None
        for command in path_commands:
            if isinstance(command, SVGCommandMove):
                # print('in move',command)
                if svg_path is not None and (allow_empty or svg_path.path_commands):  # SVGPath contains at least one command
                    if add_closing:
                        svg_path.closed = True
                    if not add_closing:
                        svg_path.path_commands.append(empty_command)
                    svg_paths.append(svg_path)

                svg_path = SVGPath([], command.start_pos.copy(), filling=filling)
            else:

                if svg_path is None:
                    # Ignore commands until the first moveTo commands
                    continue

                if isinstance(command, SVGCommandClose):
                    # print('not in move,in close',command) #Z
                    if allow_empty or svg_path.path_commands:  # SVGPath contains at least one command
                        svg_path.closed = True
                        if not svg_path.path_commands:
                            svg_path.path_commands.append(empty_command)
                        #连接起点与终点
                        if svg_path.path_commands[0].start_pos!=svg_path.path_commands[-1].end_pos:
                            svg_path.path_commands.append(SVGCommandLine(
                                                                     svg_path.path_commands[-1].end_pos,svg_path.path_commands[0].start_pos))
                        svg_paths.append(svg_path)
                    svg_path = None
                else:
                    svg_path.path_commands.append(command)
        if svg_path is not None and (allow_empty or svg_path.path_commands):  # SVGPath contains at least one command
            if add_closing:
                svg_path.closed = True
            if not svg_path.path_commands:
                svg_path.path_commands.append(empty_command)
            svg_paths.append(svg_path)
        return SVGPathGroup(svg_paths, fill=fill)

    def __repr__(self):
        return "SVGPath({})".format(" ".join(command.__repr__() for command in self.all_commands()))

    def to_str(self, fill=False):
        return " ".join(command.to_str() for command in self.all_commands())

    def to_tensor(self, PAD_VAL=-1):
        return torch.stack([command.to_tensor(PAD_VAL=PAD_VAL) for command in self.all_commands()])

    def _get_viz_elements(self, with_points=False, with_handles=False, with_bboxes=False, color_firstlast=False, with_moves=True):
        points = self._get_points_viz(color_firstlast, with_moves) if with_points else ()
        handles = self._get_handles_viz() if with_handles else ()
        return [*points, *handles]

    def draw(self, viewbox=Bbox(24), *args, **kwargs):
        from .svg import SVG
        return SVG([self.to_group()], viewbox=viewbox).draw(*args, **kwargs)

    def save_svg(self, viewbox=Bbox(24), *args, **kwargs):
        from .svg import SVG
        return SVG([self.to_group()], viewbox=viewbox).save_svg(*args, **kwargs)

    def _get_points_viz(self, color_firstlast=True, with_moves=True):
        points = []
        commands = self.all_commands(with_close=False)
        n = len(commands)
        for i, command in enumerate(commands):
            if not isinstance(command, SVGCommandMove) or with_moves:
                points_viz = command.get_points_viz(first=(color_firstlast and i <= 1), last=(color_firstlast and i >= n-2))
                points.extend(points_viz)
        return points

    def _get_handles_viz(self):
        handles = []
        for command in self.path_commands:
            handles.extend(command.get_handles_viz())
        return handles

    def _get_unique_geoms(self):
        geoms = []
        for command in self.all_commands():
            geoms.extend(command.get_geoms())
        return list(set(geoms))

    def translate(self, vec):
        for geom in self._get_unique_geoms():
            geom.translate(vec)
        return self

    def rotate(self, angle):
        for geom in self._get_unique_geoms():
            geom.rotate_(angle)
        return self

    def scale(self, factor):
        for geom in self._get_unique_geoms():
            geom.scale(factor)
        return self

    def filter_consecutives(self):
        path_commands = []
        for command in self.path_commands:
            if not command.start_pos.isclose(command.end_pos):
                path_commands.append(command)
        self.path_commands = path_commands
        return self

    def filter_duplicates(self, min_dist=0.2):
        path_commands = []
        current_command = None
        for command in self.path_commands:
            if current_command is None:
                path_commands.append(command)
                current_command = command

            if command.end_pos.dist(current_command.end_pos) >= min_dist:
                command.start_pos = current_command.end_pos
                path_commands.append(command)
                current_command = command

        self.path_commands = path_commands
        return self

    def duplicate_extremities(self):
        self.path_commands = [SVGCommandLine(self.start_pos, self.start_pos),
                              *self.path_commands,
                              SVGCommandLine(self.end_pos, self.end_pos)]
        # print(self.path_commands,'in duplicate_extremities')
        return self

    def is_clockwise(self):
        if len(self.path_commands) == 1:
            cmd = self.path_commands[0]
            return cmd.start_pos.tolist() <= cmd.end_pos.tolist()

        det_total = 0.
        for cmd in self.path_commands:
            det_total += geom.det(cmd.start_pos, cmd.end_pos)
        return det_total >= 0.

    def set_orientation(self, orientation):
        """
        orientation: 1 (clockwise), 0 (counter-clockwise)
        """
        if orientation == self.is_clockwise():
            return self
        return self.reverse()

    def set_closed(self, closed=True):
        self.closed = closed
        return self

    def reverse(self):
        path_commands = []

        for command in reversed(self.path_commands):
            path_commands.append(command.reverse())

        self.path_commands = path_commands
        return self

    def reverse_non_closed(self):
        if not self.start_pos.isclose(self.end_pos):
            return self.reverse()
        return self

    def simplify_arcs(self):
        path_commands = []
        for command in self.path_commands:
            if isinstance(command, SVGCommandArc):
                if command.radius.iszero():
                    continue
                if command.start_pos.isclose(command.end_pos):
                    continue
                path_commands.extend(command.to_beziers())
            else:
                path_commands.append(command)

        self.path_commands = path_commands
        return self

    def _get_topleftmost_command(self):
        topleftmost_cmd = None
        topleftmost_idx = 0

        for i, cmd in enumerate(self.path_commands):
            if topleftmost_cmd is None or cmd.is_left_to(topleftmost_cmd):
                topleftmost_cmd = cmd
                topleftmost_idx = i

        return topleftmost_cmd, topleftmost_idx

    def reorder(self):
        if self.closed:
            topleftmost_cmd, topleftmost_idx = self._get_topleftmost_command()
            self.path_commands = [
                *self.path_commands[topleftmost_idx:],
                *self.path_commands[:topleftmost_idx]
            ]

        return self

    def to_video(self, wrapper, clips=None, svg_commands=None, color="grey"):
        from .svg import SVG
        from .svg_primitive import SVGLine, SVGCircle

        if clips is None:
            clips = []
        if svg_commands is None:
            svg_commands = []
        svg_dots, svg_moves = [], []

        for command in self.all_commands():
            start_pos, end_pos = command.start_pos, command.end_pos

            if isinstance(command, SVGCommandMove):
                move = SVGLine(start_pos, end_pos, color="teal", dasharray=0.5)
                svg_moves.append(move)

            dot = SVGCircle(end_pos, radius=Radius(0.1), color="red")
            svg_dots.append(dot)

            svg_path = SVGPath(svg_commands).to_group(color=color)
            svg_new_path = SVGPath([SVGCommandMove(start_pos), command]).to_group(color="red")

            svg_paths = [svg_path, svg_new_path]  if svg_commands else [svg_new_path]
            im = SVG([*svg_paths, *svg_moves, *svg_dots]).draw(do_display=False, return_png=True, with_points=False)
            clips.append(wrapper(np.array(im)))

            svg_dots[-1].color = "grey"
            svg_commands.append(command)
            svg_moves = []

        return clips, svg_commands

    def numericalize(self, n=256):
        for command in self.all_commands():
            command.numericalize(n)

    def smooth(self):
        # https://github.com/paperjs/paper.js/blob/c7d85b663edb728ec78fffa9f828435eaf78d9c9/src/path/Path.js#L1288
        n = len(self.path_commands)
        knots = [self.start_pos, *(path_commmand.end_pos for path_commmand in self.path_commands)]
        r = [knots[0] + 2 * knots[1]]
        f = [2]
        p = [Point(0.)] * (n + 1)

        # Solve with the Thomas algorithm
        for i in range(1, n):
            internal = i < n - 1
            a = 1
            b = 4 if internal else 2
            u = 4 if internal else 3
            v = 2 if internal else 0
            m = a / f[i-1]

            f.append(b-m)
            r.append(u * knots[i] + v * knots[i + 1] - m * r[i-1])

        p[n-1] = r[n-1] / f[n-1]
        for i in range(n-2, -1, -1):
            p[i] = (r[i] - p[i+1]) / f[i]
        p[n] = (3 * knots[n] - p[n-1]) / 2

        for i in range(n):
            p1, p2 = knots[i], knots[i+1]
            c1, c2 = p[i], 2 * p2 - p[i+1]
            self.path_commands[i] = SVGCommandBezier(p1, c1, c2, p2)

        return self

    def simplify_heuristic(self):
        return self.copy().split(max_dist=2, include_lines=False) \
            .simplify(tolerance=0.1, epsilon=0.2, angle_threshold=150) \
            .split(max_dist=7.5)

    def simplify(self, tolerance=0.1, epsilon=0.1, angle_threshold=179., force_smooth=False):
        # https://github.com/paperjs/paper.js/blob/c044b698c6b224c10a7747664b2a4cd00a416a25/src/path/PathFitter.js#L44
        points = [self.start_pos, *(path_command.end_pos for path_command in self.path_commands)]

        def subdivide_indices():
            segments_list = []
            current_segment = []
            prev_command = None

            for i, command in enumerate(self.path_commands):
                if isinstance(command, SVGCommandLine):
                    if current_segment:
                        segments_list.append(current_segment)
                        current_segment = []
                    prev_command = None

                    continue

                if prev_command is not None and prev_command.angle(command) < angle_threshold:
                    if current_segment:
                        segments_list.append(current_segment)
                        current_segment = []

                current_segment.append(i)
                prev_command = command

            if current_segment:
                segments_list.append(current_segment)

            return segments_list

        path_commands = []

        def computeMaxError(first, last, curve: SVGCommandBezier, u):
            maxDist = 0.
            index = (last - first + 1) // 2
            for i in range(1, last - first):
                dist = curve.eval(u[i]).dist(points[first + i]) ** 2
                if dist >= maxDist:
                    maxDist = dist
                    index = first + i
            return maxDist, index

        def chordLengthParametrize(first, last):
            u = [0.]
            for i in range(1, last - first + 1):
                u.append(u[i-1] + points[first + i].dist(points[first + i-1]))

            for i, _ in enumerate(u[1:], 1):
                u[i] /= u[-1]

            return u

        def isMachineZero(val):
            MACHINE_EPSILON = 1.12e-16
            return val >= -MACHINE_EPSILON and val <= MACHINE_EPSILON

        def findRoot(curve: SVGCommandBezier, point, u):
            """
               Newton's root finding algorithm calculates f(x)=0 by reiterating
               x_n+1 = x_n - f(x_n)/f'(x_n)
               We are trying to find curve parameter u for some point p that minimizes
               the distance from that point to the curve. Distance point to curve is d=q(u)-p.
               At minimum distance the point is perpendicular to the curve.
               We are solving
               f = q(u)-p * q'(u) = 0
               with
               f' = q'(u) * q'(u) + q(u)-p * q''(u)
               gives
               u_n+1 = u_n - |q(u_n)-p * q'(u_n)| / |q'(u_n)**2 + q(u_n)-p * q''(u_n)|
            """
            diff = curve.eval(u) - point
            d1, d2 = curve.derivative(u, n=1), curve.derivative(u, n=2)
            numerator = diff.dot(d1)
            denominator = d1.dot(d1) + diff.dot(d2)

            return u if isMachineZero(denominator) else u - numerator / denominator

        def reparametrize(first, last, u, curve: SVGCommandBezier):
            for i in range(0, last - first + 1):
                u[i] = findRoot(curve, points[first + i], u[i])

            for i in range(1, len(u)):
                if u[i] <= u[i-1]:
                    return False

            return True

        def generateBezier(first, last, uPrime, tan1, tan2):
            epsilon = 1e-12
            p1, p2 = points[first], points[last]
            C = np.zeros((2, 2))
            X = np.zeros(2)

            for i in range(last - first + 1):
                u = uPrime[i]
                t = 1 - u
                b = 3 * u * t
                b0 = t**3
                b1 = b * t
                b2 = b * u
                b3 = u**3
                a1 = tan1 * b1
                a2 = tan2 * b2
                tmp = points[first + i] - p1 * (b0 + b1) - p2 * (b2 + b3)

                C[0, 0] += a1.dot(a1)
                C[0, 1] += a1.dot(a2)
                C[1, 0] = C[0, 1]
                C[1, 1] += a2.dot(a2)
                X[0] += a1.dot(tmp)
                X[1] += a2.dot(tmp)

            detC0C1 = C[0, 0] * C[1, 1] - C[1, 0] * C[0, 1]
            if abs(detC0C1) > epsilon:
                detC0X = C[0, 0] * X[1] - C[1, 0] * X[0]
                detXC1 = X[0] * C[1, 1] - X[1] * C[0, 1]
                alpha1 = detXC1 / detC0C1
                alpha2 = detC0X / detC0C1
            else:
                c0 = C[0, 0] + C[0, 1]
                c1 = C[1, 0] + C[1, 1]
                alpha1 = alpha2 = X[0] / c0 if abs(c0) > epsilon else (X[1] / c1 if abs(c1) > epsilon else 0)

            segLength = p2.dist(p1)
            eps = epsilon * segLength
            handle1 = handle2 = None

            if alpha1 < eps or alpha2 < eps:
                alpha1 = alpha2 = segLength / 3
            else:
                line = p2 - p1
                handle1 = tan1 * alpha1
                handle2 = tan2 * alpha2

                if handle1.dot(line) - handle2.dot(line) > segLength**2:
                    alpha1 = alpha2 = segLength / 3
                    handle1 = handle2 = None

            if handle1 is None or handle2 is None:
                handle1 = tan1 * alpha1
                handle2 = tan2 * alpha2

            return SVGCommandBezier(p1, p1 + handle1, p2 + handle2, p2)

        def computeLinearMaxError(first, last):
            maxDist = 0.
            index = (last - first + 1) // 2

            p1, p2 = points[first], points[last]
            for i in range(first + 1, last):
                dist = points[i].distToLine(p1, p2)
                if dist >= maxDist:
                    maxDist = dist
                    index = i
            return maxDist, index

        def ramerDouglasPeucker(first, last, epsilon):
            max_error, split_index = computeLinearMaxError(first, last)

            if max_error > epsilon:
                ramerDouglasPeucker(first, split_index, epsilon)
                ramerDouglasPeucker(split_index, last, epsilon)
            else:
                p1, p2 = points[first], points[last]
                path_commands.append(SVGCommandLine(p1, p2))

        def fitCubic(error, first, last, tan1=None, tan2=None):
            # For convenience, compute extremity tangents if not provided
            if tan1 is None and tan2 is None:
                tan1 = (points[first + 1] - points[first]).normalize()
                tan2 = (points[last - 1] - points[last]).normalize()

            if last - first == 1:
                p1, p2 = points[first], points[last]
                dist = p1.dist(p2) / 3
                path_commands.append(SVGCommandBezier(p1, p1 + dist * tan1, p2 + dist * tan2, p2))
                return

            uPrime = chordLengthParametrize(first, last)
            maxError = max(error, error**2)
            parametersInOrder = True

            for i in range(5):
                curve = generateBezier(first, last, uPrime, tan1, tan2)

                max_error, split_index = computeMaxError(first, last, curve, uPrime)

                if max_error < error and parametersInOrder:
                    path_commands.append(curve)
                    return

                if max_error >= maxError:
                    break

                parametersInOrder = reparametrize(first, last, uPrime, curve)
                maxError = max_error

            tanCenter = (points[split_index-1] - points[split_index+1]).normalize()
            fitCubic(error, first, split_index, tan1, tanCenter)
            fitCubic(error, split_index, last, -tanCenter, tan2)

        segments_list = subdivide_indices()
        if force_smooth:
            fitCubic(tolerance, 0, len(points) - 1)
        else:
            if segments_list:
                seg = segments_list[0]
                ramerDouglasPeucker(0, seg[0], epsilon)

                for seg, seg_next in zip(segments_list[:-1], segments_list[1:]):
                    fitCubic(tolerance, seg[0], seg[-1] + 1)
                    ramerDouglasPeucker(seg[-1] + 1, seg_next[0], epsilon)

                seg = segments_list[-1]
                fitCubic(tolerance, seg[0], seg[-1] + 1)
                ramerDouglasPeucker(seg[-1] + 1, len(points) - 1, epsilon)
            else:
                ramerDouglasPeucker(0, len(points) - 1, epsilon)

        self.path_commands = path_commands

        return self

    def split(self, n=None, max_dist=None, include_lines=True):
        path_commands = []

        for command in self.path_commands:
            if isinstance(command, SVGCommandLine) and not include_lines:
                path_commands.append(command)
            else:
                l = command.length()
                if max_dist is not None:
                    n = max(math.ceil(l / max_dist), 1)

                path_commands.extend(command.split(n=n))

        self.path_commands = path_commands

        return self

    def bbox(self):
        return union_bbox([cmd.bbox() for cmd in self.path_commands])

    def sample_points(self, max_dist=0.4):
        points = []

        for command in self.path_commands:
            l = command.length()
            n = max(math.ceil(l / max_dist), 1)
            points.extend(command.sample_points(n=n, return_array=True)[None])
        points = np.concatenate(points, axis=0)
        return points

    def to_shapely(self):
        polygon = shapely.geometry.Polygon(self.sample_points())

        if not polygon.is_valid:
            polygon = polygon.buffer(0)

        return polygon

    def to_points(self):
        return np.array([self.start_pos.pos, *(cmd.end_pos.pos for cmd in self.path_commands)])



import shutil
import torch
import torch.nn as nn
import os
import random
import numpy as np
import glob


def save_ckpt(checkpoint_dir, model, cfg=None, optimizer=None, scheduler_lr=None, scheduler_warmup=None,
              stats=None, train_vars=None):
    if is_multi_gpu(model):
        model = model.module

    state = {
        "model": model.state_dict()
    }

    if optimizer is not None:
        state["optimizer"] = optimizer.state_dict()
    if scheduler_lr is not None:
        state["scheduler_lr"] = scheduler_lr.state_dict()
    if scheduler_warmup is not None:
        state["scheduler_warmup"] = scheduler_warmup.state_dict()
    if cfg is not None:
        state["cfg"] = cfg.to_dict()
    if stats is not None:
        state["stats"] = stats.to_dict()
    if train_vars is not None:
        state["train_vars"] = train_vars.to_dict()

    checkpoint_path = os.path.join(checkpoint_dir, "{:06d}.pth.tar".format(stats.step))

    if not os.path.exists(checkpoint_dir):
        os.makedirs(checkpoint_dir)
    torch.save(state, checkpoint_path)

    if stats.is_best():
        best_model_path = os.path.join(checkpoint_dir, "best.pth.tar")
        shutil.copyfile(checkpoint_path, best_model_path)


def save_ckpt_list(checkpoint_dir, model, cfg=None, optimizers=None, scheduler_lrs=None, scheduler_warmups=None,
              stats=None, train_vars=None):
    if is_multi_gpu(model):
        model = model.module

    state = {
        "model": model.state_dict()
    }

    if optimizers is not None:
        state["optimizers"] = [optimizer.state_dict() if optimizer is not None else optimizer for optimizer in optimizers]
    if scheduler_lrs is not None:
        state["scheduler_lrs"] = [scheduler_lr.state_dict() if scheduler_lr is not None else scheduler_lr for scheduler_lr in scheduler_lrs]
    if scheduler_warmups is not None:
        state["scheduler_warmups"] = [scheduler_warmup.state_dict() if scheduler_warmup is not None else None for scheduler_warmup in scheduler_warmups]
    if cfg is not None:
        state["cfg"] = cfg.to_dict()
    if stats is not None:
        state["stats"] = stats.to_dict()
    if train_vars is not None:
        state["train_vars"] = train_vars.to_dict()

    checkpoint_path = os.path.join(checkpoint_dir, "{:06d}.pth.tar".format(stats.step))

    if not os.path.exists(checkpoint_dir):
        os.makedirs(checkpoint_dir)
    torch.save(state, checkpoint_path)

    if stats.is_best():
        best_model_path = os.path.join(checkpoint_dir, "best.pth.tar")
        shutil.copyfile(checkpoint_path, best_model_path)


def load_ckpt(checkpoint_dir, model, cfg=None, optimizer=None, scheduler_lr=None, scheduler_warmup=None,
              stats=None, train_vars=None):
    if not os.path.exists(checkpoint_dir):
        return False

    if os.path.isfile(checkpoint_dir):
        checkpoint_path = checkpoint_dir
    else:
        ckpts_paths = sorted(glob.glob(os.path.join(checkpoint_dir, "./[0-9]*.pth.tar")))
        if not ckpts_paths:
            return False
        checkpoint_path = ckpts_paths[-1]

    state = torch.load(checkpoint_path)

    if is_multi_gpu(model):
        model = model.module
    model.load_state_dict(state["model"], strict=False)

    if optimizer is not None:
        optimizer.load_state_dict(state["optimizer"])
    if scheduler_lr is not None:
        scheduler_lr.load_state_dict(state["scheduler_lr"])
    if scheduler_warmup is not None:
        scheduler_warmup.load_state_dict(state["scheduler_warmup"])
    if cfg is not None:
        cfg.load_dict(state["cfg"])
    if stats is not None:
        stats.load_dict(state["stats"])
    if train_vars is not None:
        train_vars.load_dict(state["train_vars"])

    return True


def load_ckpt_list(checkpoint_dir, model, cfg=None, optimizers=None, scheduler_lrs=None, scheduler_warmups=None,
              stats=None, train_vars=None):
    if not os.path.exists(checkpoint_dir):
        return False

    if os.path.isfile(checkpoint_dir):
        checkpoint_path = checkpoint_dir
    else:
        ckpts_paths = sorted(glob.glob(os.path.join(checkpoint_dir, "./[0-9]*.pth.tar")))
        if not ckpts_paths:
            return False
        checkpoint_path = ckpts_paths[-1]

    state = torch.load(checkpoint_path)

    if is_multi_gpu(model):
        model = model.module
    model.load_state_dict(state["model"], strict=False)

    for optimizer, scheduler_lr, scheduler_warmup, optimizer_sd, scheduler_lr_sd, scheduler_warmups_sd in zip(optimizers, scheduler_lrs, scheduler_warmups, state["optimizers"], state["scheduler_lrs"], state["scheduler_warmups"]):
        if optimizer is not None and optimizer_sd is not None:
            optimizer.load_state_dict(optimizer_sd)
        if scheduler_lr is not None and scheduler_lr_sd is not None:
            scheduler_lr.load_state_dict(scheduler_lr_sd)
        if scheduler_warmup is not None and scheduler_warmups_sd is not None:
            scheduler_warmup.load_state_dict(scheduler_warmups_sd)
    if cfg is not None and state["cfg"] is not None:
        cfg.load_dict(state["cfg"])
    if stats is not None and state["stats"] is not None:
        stats.load_dict(state["stats"])
    if train_vars is not None and state["train_vars"] is not None:
        train_vars.load_dict(state["train_vars"])

    return True


def load_model(checkpoint_path, model):
    state = torch.load(checkpoint_path)

    if is_multi_gpu(model):
        model = model.module
    model.load_state_dict(state["model"], strict=False)


def is_multi_gpu(model):
    return isinstance(model, nn.DataParallel)


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def pad_sequence(sequences, batch_first=False, padding_value=0, max_len=None):
    r"""Pad a list of variable length Tensors with ``padding_value``

    ``pad_sequence`` stacks a list of Tensors along a new dimension,
    and pads them to equal length. For example, if the input is list of
    sequences with size ``L x *`` and if batch_first is False, and ``T x B x *``
    otherwise.

    `B` is batch size. It is equal to the number of elements in ``sequences``.
    `T` is length of the longest sequence.
    `L` is length of the sequence.
    `*` is any number of trailing dimensions, including none.

    Example:
        >>> from torch.nn.utils.rnn import pad_sequence
        >>> a = torch.ones(25, 300)
        >>> b = torch.ones(22, 300)
        >>> c = torch.ones(15, 300)
        >>> pad_sequence([a, b, c]).size()
        torch.Size([25, 3, 300])

    Note:
        This function returns a Tensor of size ``T x B x *`` or ``B x T x *``
        where `T` is the length of the longest sequence. This function assumes
        trailing dimensions and type of all the Tensors in sequences are same.

    Arguments:
        sequences (list[Tensor]): list of variable length sequences.
        batch_first (bool, optional): output will be in ``B x T x *`` if True, or in
            ``T x B x *`` otherwise
        padding_value (float, optional): value for padded elements. Default: 0.

    Returns:
        Tensor of size ``T x B x *`` if :attr:`batch_first` is ``False``.
        Tensor of size ``B x T x *`` otherwise
    """

    # assuming trailing dimensions and type of all the Tensors
    # in sequences are same and fetching those from sequences[0]
    max_size = sequences[0].size()
    trailing_dims = max_size[1:]

    if max_len is None:
        max_len = max([s.size(0) for s in sequences])
    if batch_first:
        out_dims = (len(sequences), max_len) + trailing_dims
    else:
        out_dims = (max_len, len(sequences)) + trailing_dims

    out_tensor = sequences[0].data.new(*out_dims).fill_(padding_value)
    for i, tensor in enumerate(sequences):
        length = tensor.size(0)
        # use index notation to prevent duplicate references to the tensor
        if batch_first:
            out_tensor[i, :length, ...] = tensor
        else:
            out_tensor[:length, i, ...] = tensor

    return out_tensor


def set_seed(_seed=42):
    random.seed(_seed)
    np.random.seed(_seed)
    torch.manual_seed(_seed)
    torch.cuda.manual_seed(_seed)
    torch.cuda.manual_seed_all(_seed)
    os.environ['PYTHONHASHSEED'] = str(_seed)


def infinite_range(start_idx=0):
    while True:
        yield start_idx
        start_idx += 1



from kivy.app import App
from kivy.uix.widget import Widget
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.scatter import Scatter
from kivy.uix.label import Label
from kivy.uix.scrollview import ScrollView
from kivy.properties import BooleanProperty, StringProperty, NumericProperty, ListProperty, ObjectProperty
from kivy.uix.behaviors import ButtonBehavior
from kivy.vector import Vector
from kivy.metrics import dp
from kivy.clock import Clock
from kivy.uix.popup import Popup

from kivy.config import Config
Config.set('graphics', 'width', '1400')
Config.set('graphics', 'height', '800')
from kivy.core.window import Window

import os
from typing import List

from deepsvg.svglib.geom import Point
from deepsvg.svglib.svg_command import SVGCommandMove, SVGCommandLine, SVGCommandBezier
from deepsvg.svgtensor_dataset import SVGTensorDataset

from .layout.aligned_textinput import AlignedTextInput
from .state.state import State, ToolMode, DrawMode, LoopMode, PlaybackMode
from .state.project import Frame
from .config import ROOT_DIR
from .interpolate import compute_interpolation
from .utils import *


if not os.path.exists(ROOT_DIR):
    os.makedirs(ROOT_DIR)

state = State()
state.load_state()
state.load_project()


class HeaderIcon(Button):
    index = NumericProperty(0)
    source = StringProperty("")

    def on_press(self):
        state.header.selected_tool = self.index


class Header(BoxLayout):
    selected_tool = NumericProperty(0)
    title = StringProperty(state.project.name)
    is_playing = BooleanProperty(False)
    delay = NumericProperty(state.delay)

    def on_selected_tool(self, *args):
        if self.selected_tool in [ToolMode.MOVE, ToolMode.PEN, ToolMode.PENCIL] and state.header.is_playing:
            state.header.pause_animation()

    def on_done(self, *args):
        if self.selected_tool == ToolMode.PEN and state.draw_mode == DrawMode.DRAW:
            path = state.current_path

            last_segment = path.children[-1]
            path.remove_widget(last_segment)

            state.draw_viewbox.on_path_done(state.current_path)

            state.draw_mode = DrawMode.STILL
            state.current_path = None
            self.selected_tool = ToolMode.MOVE

    def on_erase(self):
        state.modified = True
        state.draw_viewbox.clear()

        state.timeline.make_keyframe(False)

    def add_frame(self, keyframe=False):
        frame_idx = state.timeline._add_frame(keyframe=keyframe)

        state.project.frames.append(Frame(frame_idx, keyframe))

        self.load_next_frame(frame_idx=frame_idx)

    def play_animation(self):
        self.is_playing = True
        state.sidebar.selected_path_idx = -1
        self.clock = Clock.schedule_once(self.load_next_frame)

    def load_next_frame(self, dt=0, frame_idx=None, *args):
        if state.timeline.nb_frames > 0:
            if frame_idx is None:
                frame_idx_tmp = state.timeline.selected_frame + state.loop_orientation

                if frame_idx_tmp < 0 or frame_idx_tmp >= state.timeline.nb_frames:
                    if state.loop_mode in [LoopMode.NORMAL, LoopMode.REVERSE]:
                        frame_idx = frame_idx_tmp % state.timeline.nb_frames
                    else:  # LoopMode.PINGPONG
                        state.loop_orientation *= -1
                        frame_idx = (state.timeline.selected_frame + state.loop_orientation) % state.timeline.nb_frames
                else:
                    frame_idx = frame_idx_tmp

            state.timeline.selected_frame = frame_idx

            if self.is_playing:
                if state.playback_mode == PlaybackMode.EASE:
                    t = frame_idx / state.timeline.nb_frames
                    delay = 2 * state.delay / (1 + d_easein_easeout(t))
                else:
                    delay = state.delay
                self.clock = Clock.schedule_once(self.load_next_frame, delay)

    def pause_animation(self):
        self.clock.cancel()
        state.sidebar.selected_path_idx = -1
        self.is_playing = False

        state.timeline.on_selected_frame()  # re-render frame to display sidebar layers

    def on_title(self, title):
        state.project.name = title

    def interpolate(self):
        state.draw_viewbox.save_frame()

        compute_interpolation(state.project)


class PathLayerView(ButtonBehavior, BoxLayout):
    index = NumericProperty(0)
    source = StringProperty("")

    def __init__(self, index, **kwargs):
        super().__init__(**kwargs)

        self.index = index
        self.source = os.path.join(state.project.cache_dir, f"{state.timeline.selected_frame}_{index}.png")

    def on_press(self):
        state.sidebar.selected_path_idx = self.index

    def move_up(self):
        if self.index > 0:
            state.sidebar.swap_paths(self.index, self.index - 1)

    def move_down(self):
        if self.index < state.sidebar.nb_paths - 1:
            state.sidebar.swap_paths(self.index, self.index + 1)

    def reverse(self):
        state.sidebar.reverse_path(self.index)


class Sidebar(ScrollView):
    selected_path_idx = NumericProperty(-1)

    @property
    def sidebar(self):
        return self.ids.sidebar

    @property
    def nb_paths(self):
        return len(self.sidebar.children)

    def on_selected_path_idx(self, *args):
        state.draw_viewbox.unselect_all()

        if self.selected_path_idx >= 0:
            state.draw_viewbox.get_path(self.selected_path_idx).selected = True

    def _add_path(self, idx=None):
        if idx is None:
            idx = self.nb_paths
        new_pathlayer = PathLayerView(idx)
        self.sidebar.add_widget(new_pathlayer)
        return idx

    def get_path(self, path_idx):
        index = self.nb_paths - 1 - path_idx
        return self.sidebar.children[index]

    def erase(self):
        self.sidebar.clear_widgets()
        self.selected_path_idx = -1

    def swap_paths(self, idx1, idx2):
        path_layer1, path_layer2 = self.get_path(idx1), self.get_path(idx2)
        path1, path2 = state.draw_viewbox.get_path(idx1), state.draw_viewbox.get_path(idx2)

        path_layer1.index, path_layer2.index = idx2, idx1
        path1.color, path2.color = path2.color, path1.color
        path1.index, path2.index = path2.index, path1.index

        id1, id2 = self.nb_paths - 1 - idx1, self.nb_paths - 1 - idx2
        self.sidebar.children[id1], self.sidebar.children[id2] = path_layer2, path_layer1
        state.draw_viewbox.children[id1], state.draw_viewbox.children[id2] = path2, path1

        self.selected_path_idx = idx2
        state.modified = True

    def reverse_path(self, idx):
        path = state.draw_viewbox.get_path(idx)
        svg_path = path.to_svg_path().reverse()
        new_path = BezierPath.from_svg_path(svg_path, color=path.color, index=path.index, selected=path.selected)

        id = self.nb_paths - 1 - idx
        state.draw_viewbox.remove_widget(path)
        state.draw_viewbox.add_widget(new_path, index=id)

        self.selected_path_idx = idx
        state.modified = True

    def select(self, path_idx):
        if self.selected_path_idx >= 0:
            state.draw_viewbox.get_path(state.sidebar.selected_path_idx).selected = False
        self.selected_path_idx = path_idx


class BezierSegment(Widget):
    is_curved = BooleanProperty(True)

    is_finished = BooleanProperty(True)
    select_dist = NumericProperty(3)

    p1 = ListProperty([0, 0])
    q1 = ListProperty([0, 0])
    q2 = ListProperty([0, 0])
    p2 = ListProperty([0, 0])

    def clone(self):
        segment = BezierSegment()
        segment.is_curved = self.is_curved
        segment.p1 = self.p1  # shallow copy
        segment.q1 = self.q1
        segment.q2 = self.q2
        segment.p2 = self.p2
        return segment

    @staticmethod
    def line(p1, p2):
        segment = BezierSegment()
        segment.is_curved = False
        segment.p1 = segment.q1 = p1
        segment.p2 = segment.q2 = p2
        return segment

    @staticmethod
    def bezier(p1, q1, q2, p2):
        segment = BezierSegment()
        segment.is_curved = True
        segment.q1, segment.q2 = q1, q2
        segment.p1, segment.p2 = p1, p2
        return segment

    def get_point(self, key):
        return getattr(self, key)

    def on_touch_down(self, touch):
        max_dist = dp(self.select_dist)
        
        if not self.parent.selected:
            return super().on_touch_down(touch)

        keys_to_test = ["p1", "q1", "q2", "p2"] if self.is_curved else ["p1", "p2"]
        for key in keys_to_test:
            if dist(touch.pos, getattr(self, key)) < max_dist:
                touch.ud['selected'] = key
                touch.grab(self)

                state.modified = True

                return True

    def on_touch_move(self, touch):
        if touch.grab_current is not self:
            return super().on_touch_move(touch)

        key = touch.ud['selected']
        setattr(self, key, touch.pos)

        if state.header.selected_tool == ToolMode.PEN:
            self.is_curved = True
            self.is_finished = False
            state.draw_mode = DrawMode.HOLDING_DOWN

            setattr(self, "p2", touch.pos)

        if key in ["p1", "p2"]:
            self.parent.move(self, key, touch.pos)

    def on_touch_up(self, touch):
        if touch.grab_current is not self:
            return super().on_touch_up(touch)

        touch.ungrab(self)

        if state.header.selected_tool == ToolMode.PEN:
            self.is_finished = True
            state.draw_mode = DrawMode.DRAW


class BezierPath(Widget):
    color = ListProperty([1, 1, 1])
    index = NumericProperty(0)
    selected = BooleanProperty(False)

    def __init__(self, segments: List[BezierSegment], color=None, index=None, selected=False, **kwargs):
        super().__init__(**kwargs)

        if color is not None:
            self.color = color

        if index is not None:
            self.index = index

        self.selected = selected

        for segment in segments:
            self.add_segment(segment)

    def clone(self):
        segments = [segment.clone() for segment in self.children]
        return BezierPath(segments, self.color, self.index, self.selected)

    def add_segment(self, segment: BezierSegment):
        self.add_widget(segment, index=len(self.children))

    def move(self, segment, key, pos):
        idx = self.children.index(segment)

        if not (idx == 0 and key == "p1") and not (idx == len(self.children) - 1 and key == "p2"):
            idx2, key2 = (idx-1, "p2") if key == "p1" else (idx+1, "p1")
            setattr(self.children[idx2], key2, pos)

    def add_widget(self, widget, index=0, canvas=None):
        super().add_widget(widget, index=index, canvas=canvas)

    def remove_widget(self, widget):
        super().remove_widget(widget)

    @staticmethod
    def from_svg_path(svg_path: SVGPath, *args, **kwargs):
        segments = []
        for command in svg_path.path_commands:
            if isinstance(command, SVGCommandBezier):
                segment = BezierSegment.bezier(flip_vertical(command.p1.tolist()), flip_vertical(command.q1.tolist()),
                                               flip_vertical(command.q2.tolist()), flip_vertical(command.p2.tolist()))
                segments.append(segment)
            elif isinstance(command, SVGCommandLine):
                segment = BezierSegment.line(flip_vertical(command.start_pos.tolist()),
                                             flip_vertical(command.end_pos.tolist()))
                segments.append(segment)

        path = BezierPath(segments, *args, **kwargs)
        return path

    def to_svg_path(self):
        path_commands = []
        for segment in self.children:
            if segment.is_curved:
                command = SVGCommandBezier(Point(*flip_vertical(segment.p1)), Point(*flip_vertical(segment.q1)),
                                           Point(*flip_vertical(segment.q2)), Point(*flip_vertical(segment.p2)))
            else:
                command = SVGCommandLine(Point(*flip_vertical(segment.p1)), Point(*flip_vertical(segment.p2)))
            path_commands.append(command)
        # print(path_commands,'in main')
        svg_path = SVGPath(path_commands)
        return svg_path


class Sketch(Widget):
    color = ListProperty([1, 1, 1])
    points = ListProperty([])

    def __init__(self, points, color=None, **kwargs):
        super().__init__(**kwargs)

        if color is not None:
            self.color = color

        self.points = points

    def on_touch_move(self, touch):
        if touch.grab_current is not self:
            return super().on_touch_move(touch)

        self.points.extend(touch.pos)

    def on_touch_up(self, touch):
        if touch.grab_current is not self:
            return super().on_touch_up(touch)

        touch.ungrab(self)

        self.parent.on_sketch_done(self)

    def to_svg_path(self):
        points = [Point(x, 255 - y) for x, y in zip(self.points[::2], self.points[1::2])]
        commands = [SVGCommandMove(points[0])] + [SVGCommandLine(p1, p2) for p1, p2 in zip(points[:-1], points[1:])]
        svg_path = SVGPath.from_commands(commands).path
        return svg_path


class EditorView(Scatter):
    def on_touch_down(self, touch):
        if self.collide_point(*touch.pos) and touch.is_mouse_scrolling:
            if touch.button == 'scrolldown':
                if self.scale < 10:
                    self.scale = self.scale * 1.1
            elif touch.button == 'scrollup':
                if self.scale > 1:
                    self.scale = self.scale * 0.8
            return True

        return super().on_touch_down(touch)


class DrawViewbox(Widget):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Window.bind(mouse_pos=self.on_mouse_pos)

    @property
    def nb_paths(self):
        return len(self.children)

    def _get_color(self, idx):
        color = color_dict[colors[idx % len(colors)]]
        return color

    def on_mouse_pos(self, _, abs_pos):
        pos = (Vector(abs_pos) - Vector(self.parent.pos)) / self.parent.scale

        if state.header.selected_tool == ToolMode.PEN and state.draw_mode == DrawMode.DRAW:
            segment = state.current_path.children[-1]
            segment.p2 = segment.q2 = pos

    def on_sketch_done(self, sketch: Sketch):
        # Digitalize points to Bézier path
        svg_path = preprocess_svg_path(sketch.to_svg_path(), force_smooth=True)

        path_idx = state.sidebar.nb_paths
        path = BezierPath.from_svg_path(svg_path, color=sketch.color, index=path_idx, selected=True)
        self.remove_widget(sketch)

        self.add_new_path(path, svg_path)

    def on_path_done(self, path: BezierPath):
        svg_path = preprocess_svg_path(path.to_svg_path())

        path_idx = state.sidebar.nb_paths
        new_path = BezierPath.from_svg_path(svg_path, color=path.color, index=path_idx, selected=True)
        self.remove_widget(path)

        self.add_new_path(new_path, svg_path)

    def paste(self, path: BezierPath):
        path = path.clone()

        path_idx = state.sidebar.nb_paths
        path.color = self._get_color(path_idx)
        path.selected = True

        svg_path = path.to_svg_path()

        self.add_new_path(path, svg_path)

    def unselect_all(self):
        for path in self.children:
            path.selected = False

    def get_path(self, path_idx):
        index = self.nb_paths - 1 - path_idx
        return self.children[index]

    def add_new_path(self, path: BezierSegment, svg_path: SVGPath):
        self.add_path(path, svg_path, force_rerender_miniature=True)

        state.modified = True
        state.timeline.make_keyframe(True)
        state.sidebar.select(path.index)

    def add_path(self, path: BezierPath, svg_path: SVGPath, force_rerender_miniature=False):
        path_idx = state.sidebar.nb_paths
        self.add_widget(path)

        miniature_path = os.path.join(state.project.cache_dir, f"{state.timeline.selected_frame}_{path_idx}.png")
        if not os.path.exists(miniature_path) or force_rerender_miniature:
            svg_path = normalized_path(svg_path)
            svg_path.draw(viewbox=svg_path.bbox().make_square(min_size=12),
                          file_path=os.path.join(state.project.cache_dir, f"{state.timeline.selected_frame}_{path_idx}.png"),
                          do_display=False)

        if not state.header.is_playing:
            state.sidebar._add_path()

    def on_touch_down(self, touch):
        if state.header.selected_tool == ToolMode.PLAY:
            return False

        if state.header.selected_tool == ToolMode.PEN and self.collide_point(*touch.pos):
            state.draw_mode = DrawMode.DRAW

            if state.current_path is None:
                path = BezierPath([], color=self._get_color(len(self.children)), selected=True)
                self.add_widget(path)
                state.current_path = path

            l = BezierSegment.line(touch.pos, touch.pos)

            touch.ud["selected"] = "q1"
            touch.grab(l)

            state.current_path.add_segment(l)

            state.modified = True

            return True

        if state.header.selected_tool == ToolMode.PENCIL and self.collide_point(*touch.pos):
            l = Sketch([*touch.pos], color=self._get_color(len(self.children)))
            self.add_widget(l)
            touch.grab(l)

            state.modified = True

            return True

        if super().on_touch_down(touch):
            return True

    def clear(self):
        state.draw_viewbox.clear_widgets()
        state.sidebar.erase()

    def add_widget(self, widget, index=0, canvas=None):
        super().add_widget(widget, index=index, canvas=canvas)

    def remove_widget(self, widget):
        super().remove_widget(widget)

    def to_svg(self):
        svg_path_groups = []
        for path in reversed(self.children):
            svg_path_groups.append(path.to_svg_path().to_group())

        svg = SVG(svg_path_groups, viewbox=Bbox(256))
        return svg

    def load_svg(self, svg: SVG, frame_idx):
        kivy_bezierpaths = []
        for idx, svg_path in enumerate(svg.paths):
            path = BezierPath.from_svg_path(svg_path, color=self._get_color(idx), index=idx, selected=False)
            kivy_bezierpaths.append(path)
            self.add_path(path, svg_path, force_rerender_miniature=True)

        state.project.frames[frame_idx].svg = svg
        state.project.frames[frame_idx].kivy_bezierpaths = kivy_bezierpaths

    def load_cached(self, svg: SVG, kivy_bezierpaths: List[BezierPath]):
        for path, svg_path in zip(kivy_bezierpaths, svg.paths):
            self.add_path(path, svg_path)

    def load_frame(self, frame_idx):
        svg = state.project.frames[frame_idx].svg
        kivy_bezierpaths = state.project.frames[frame_idx].kivy_bezierpaths

        if kivy_bezierpaths is None:
            self.load_svg(svg, frame_idx)
        else:
            self.load_cached(svg, kivy_bezierpaths)

        self.unselect_all()

    def save_frame(self):
        svg = self.to_svg()
        state.project.frames[state.current_frame].svg = svg
        state.project.frames[state.current_frame].kivy_bezierpaths = [child for child in reversed(self.children) if isinstance(child, BezierPath)]


class HeaderButton(Button):
    pass


class UpButton(Button):
    def on_press(self):
        self.parent.move_up()


class DownButton(Button):
    def on_press(self):
        self.parent.move_down()


class ReverseButton(Button):
    def on_press(self):
        self.parent.reverse()


class FrameView(Button):
    index = NumericProperty(0)
    keyframe = BooleanProperty(False)

    def __init__(self, index, keyframe=False, **kwargs):
        super().__init__(**kwargs)

        self.index = index
        self.keyframe = keyframe

    def on_press(self):
        state.timeline.selected_frame = self.index


class TimeLine(ScrollView):
    selected_frame = NumericProperty(-1)

    @property
    def timeline(self):
        return self.ids.timeline

    @property
    def nb_frames(self):
        return len(self.timeline.children)

    def on_selected_frame(self, *args):
        self._update_frame(self.selected_frame)

    def _update_frame(self, new_frame_idx):
        if state.current_frame >= 0 and state.modified:
            state.draw_viewbox.save_frame()

        state.current_frame = new_frame_idx
        state.draw_viewbox.clear()
        state.modified = False

        state.draw_viewbox.load_frame(new_frame_idx)

    def _add_frame(self, keyframe=False):
        idx = self.nb_frames
        new_frame = FrameView(idx, keyframe=keyframe)

        self.timeline.add_widget(new_frame)
        return idx

    def get_frame(self, frame_idx):
        index = self.nb_frames - 1 - frame_idx
        return self.timeline.children[index]

    def make_keyframe(self, is_keyframe=None):
        if is_keyframe is None:
            is_keyframe = not self.get_frame(state.timeline.selected_frame).keyframe

        self.get_frame(state.timeline.selected_frame).keyframe = is_keyframe
        state.project.frames[state.timeline.selected_frame].keyframe = is_keyframe


class TitleWidget(AlignedTextInput):
    pass


class Padding(Label):
    pass


class FileChoosePopup(Popup):
    load = ObjectProperty()
    path = StringProperty(".")


class DeepSVGWidget(BoxLayout):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        state.main_widget = self
        state.header = self.ids.header
        state.sidebar = self.ids.sidebar_scroll
        state.draw_viewbox = self.ids.editor.ids.draw_viewbox
        state.timeline = self.ids.timeline_scroll

        self._load_project()

    def _load_project(self):
        for frame in state.project.frames:
            state.timeline._add_frame(keyframe=frame.keyframe)

        state.timeline.selected_frame = 0


class DeepSVGApp(App):
    def build(self):
        self.title = 'DeepSVG Editor'

        Window.bind(on_request_close=self.on_request_close)
        Window.bind(on_keyboard=self.on_keyboard)

        return DeepSVGWidget()

    def save(self):
        state.draw_viewbox.save_frame()

        state.save_state()
        state.project.save_project()

    def on_request_close(self, *args, **kwargs):
        self.save()

        self.stop()

    def on_keyboard(self, window, key, scancode, codepoint, modifier):
        CTRL_PRESSED = (modifier == ['ctrl'] or modifier == ['meta'])

        if codepoint == "h" and not CTRL_PRESSED:
            # Hand tool
            state.header.selected_tool = ToolMode.MOVE

        elif codepoint == "p" and not CTRL_PRESSED:
            # Pen tool
            state.header.selected_tool = ToolMode.PEN

        elif CTRL_PRESSED and codepoint == "p":
            # Pencil tool
            state.header.selected_tool = ToolMode.PENCIL

        elif codepoint == "k" and not CTRL_PRESSED:
            # Make keypoint
            state.timeline.make_keyframe()

        elif CTRL_PRESSED and codepoint == 'q':
            # Quit
            self.on_request_close()

        elif CTRL_PRESSED and codepoint == 'i':
            # Import
            self.file_chooser = FileChoosePopup(load=self.on_file_chosen)
            self.file_chooser.open()

        elif CTRL_PRESSED and codepoint == "e":
            # Export
            state.project.export_to_gif(loop_mode=state.loop_mode)

        elif CTRL_PRESSED and codepoint == 'c':
            # Copy
            if state.sidebar.selected_path_idx >= 0:
                state.clipboard = state.draw_viewbox.get_path(state.sidebar.selected_path_idx).clone()

        elif CTRL_PRESSED and codepoint == 'v':
            # Paste
            if isinstance(state.clipboard, BezierPath):
                state.draw_viewbox.paste(state.clipboard)

        elif CTRL_PRESSED and codepoint == 's':
            # Save
            self.save()

        elif key == Keys.SPACEBAR:
            # Play/Pause
            state.header.selected_tool = ToolMode.PLAY

            if state.header.is_playing:
                state.header.pause_animation()
            else:
                state.header.play_animation()

        elif key == Keys.LEFT:
            # Previous frame
            if state.current_frame > 0:
                state.timeline.selected_frame = state.current_frame - 1

        elif key == Keys.RIGHT:
            # Next frame
            if state.current_frame < state.timeline.nb_frames - 1:
                state.timeline.selected_frame = state.current_frame + 1

    def on_file_chosen(self, selection):
        file_path = str(selection[0])
        self.file_chooser.dismiss()

        if file_path:
            if not file_path.endswith(".svg"):
                return

            svg = SVG.load_svg(file_path)
            svg = SVGTensorDataset.simplify(svg)
            svg = SVGTensorDataset.preprocess(svg, mean=True)

            state.draw_viewbox.load_svg(svg, frame_idx=state.timeline.selected_frame)
            state.modified = True
            state.timeline.make_keyframe(True)


if __name__ == "__main__":
    DeepSVGApp().run()



from .project import DeepSVGProject
from ..config import STATE_PATH
import pickle
import os


class ToolMode:
    MOVE = 0
    PEN = 1
    PENCIL = 2
    PLAY = 3


class DrawMode:
    STILL = 0
    DRAW = 1
    HOLDING_DOWN = 2


class LoopMode:
    NORMAL = 0
    REVERSE = 1
    PINGPONG = 2


class PlaybackMode:
    NORMAL = 0
    EASE = 1


class LoopOrientation:
    FORWARD = 1
    BACKWARD = -1


class State:
    def __init__(self):
        self.project_file = None
        self.project = DeepSVGProject()

        self.loop_mode = LoopMode.PINGPONG
        self.loop_orientation = LoopOrientation.FORWARD
        self.playback_mode = PlaybackMode.EASE

        self.delay = 1 / 10.

        self.modified = False

        # Keep track of previously selected current_frame, separately from timeline's selected_frame attribute
        self.current_frame = -1

        self.current_path = None
        self.draw_mode = DrawMode.STILL

        self.clipboard = None

        # UI references
        self.main_widget = None
        self.header = None
        self.sidebar = None
        self.draw_viewbox = None
        self.timeline = None

    def save_state(self):
        with open(STATE_PATH, "wb") as f:
            state_dict = {k: v for k, v in self.__dict__.items() if k in ["project_file"]}
            pickle.dump(state_dict, f)

    def load_state(self):
        if os.path.exists(STATE_PATH):
            with open(STATE_PATH, "rb") as f:
                self.__dict__.update(pickle.load(f))

    def load_project(self):
        if self.project_file is not None:
            self.project.load_project(self.project_file)
        else:
            self.project_file = self.project.filename



'''
生成两个单路径图形之间的10个混合图形svg，
+ 以pred和target点对之间的距离分段直接移动，b样条平滑
+ 不默认两个图形的相对位置。由用户输入混合图形变换参数，lr也自由设定（0.05-0.15）。输入60x60，resize到
'''
import os
os.environ['KMP_DUPLICATE_LIB_OK'] = 'TRUE'
import os
# os.chdir("/home/xiemengni/LogoMaker/logomaker/lib/interpolation")
from interpolation.deepsvg.svglib.geom import Point, Bbox
from interpolation.deepsvg.svglib.svg import SVG
from lib.interpolation.deepsvg.svglib.svg_path import SVGPath
from lib.interpolation.deepsvg.svglib.utils import to_gif

from lib.interpolation.deepsvg.difflib.tensor import SVGTensor
from lib.interpolation.deepsvg.difflib.utils import *
from lib.interpolation.deepsvg.difflib.loss import *
from lib.interpolation.deepsvg.svglib.geom import Bbox,Point

import torch.optim as optim
import IPython.display as ipd
from moviepy.editor import ImageClip, concatenate_videoclips, ipython_display

import shutil
import cv2
import numpy as np
import glob
import matplotlib.pyplot as plt
import random
import argparse
import math
import copy


def find_bbox(points):
    return [points[:, 0].min(), points[:, 0].max(), points[:, 1].min(), points[:, 1].max()]


def max_innercircle_cp(tensors):
    # Calculate the distances to the contour
    contours = np.array(tensors.unsqueeze(1)).astype(np.int)
    print(contours.shape)
    raw_dist = np.empty(
        [int(tensors[:, 0].max() - tensors[:, 0].min()), int(tensors[:, 1].max() - tensors[:, 1].min())],
        dtype=np.float32)
    for i in range(int(tensors[:, 0].max() - tensors[:, 0].min())):
        for j in range(int(tensors[:, 1].max() - tensors[:, 1].min())):
            raw_dist[i, j] = cv2.pointPolygonTest(contours, (j, i), True)
    minVal, maxVal, _, maxDistPt = cv2.minMaxLoc(raw_dist)
    return maxDistPt


def find_center(points, mode='bbox'):
    '''
    mode in ['bbox','outercircle','innercircle']
    '''
    if mode == 'bbox':
        bbox = find_bbox(points)
        return [(bbox[1] + bbox[0]) / 2, (bbox[2] + bbox[3]) / 2]
    elif mode == 'innercircle':
        center = max_innercircle_cp(points)
        # center = [center[0] + points[:, 0].min(), center[1] + points[:, 1].min()]
        return center
    elif mode == 'outercircle':
        (x, y), radius = cv2.minEnclosingCircle(np.array(points.unsqueeze(1)).astype(np.int))
        return [x, y]
    else:
        raise ValueError("mode must in ['bbox','outercircle','innercircle']")

def sample_points(commands,control1,control2,end_pos, n=10, mode=None):
    device =commands.device

    z = torch.linspace(0, 1, n, device=device)
    Z = torch.stack([torch.ones_like(z), z, z.pow(2), z.pow(3)], dim=1)

    Q = torch.tensor([
        [[0., 0., 0., 0.],  #  "m"
         [0., 0., 0., 0.],
         [0., 0., 0., 0.],
         [0., 0., 0., 0.]],

        [[1., 0., 0., 0.],  # "l"
         [-1, 0., 0., 1.],
         [0., 0., 0., 0.],
         [0., 0., 0., 0.]],

        [[1., 0., 0., 0.],  #  "c"
         [-3, 3., 0., 0.],
         [3., -6, 3., 0.],
         [-1, 3., -3, 1.]],

        torch.zeros(4, 4),  # "a", no support yet

        torch.zeros(4, 4),  # "EOS"
        torch.zeros(4, 4),  # "SOS"
        torch.zeros(4, 4),  # "z"
    ], device=device)

    start_pos = torch.cat([
            end_pos[:-1].new_zeros(1, 2),
            end_pos[:-1]
        ])
#     print(control1.shape)
    pos = torch.cat([start_pos.unsqueeze(1),control1.unsqueeze(1),control2.unsqueeze(1),end_pos.unsqueeze(1)],dim=1)
    commands = commands.reshape(-1).long()
    inds = (commands == 1) | (commands == 2)
    commands, pos = commands[inds], pos[inds]

#     print(commands.shape,Q[commands].shape,pos.shape)
    Z_coeffs = torch.matmul(Q[commands], pos)

    # Last point being first point of next command, we drop last point except the one from the last command
    sample_points = torch.matmul(Z, Z_coeffs)
    sample_points = torch.cat([sample_points[:, :-1].reshape(-1, 2), sample_points[-1, -1].unsqueeze(0)])

    if mode=='demo':
        return sample_points, inds
    return sample_points



def spline_smooth(conts):
    CPointX = conts[:,0]
    CPointY = conts[:,1]
    n = len(CPointX)
    knotType = True
    knot = list(np.arange(n+3+1))
    for i in range(0,4):
        knot[i] = 0.0
        knot[n+i] = 1.0


    if knotType:
        L = list(np.arange(n-1))
        S = 0
        for i in range(n-1):
            L[i] = np.sqrt(pow(CPointX[i+1]- CPointX[i],2) + pow(CPointY[i+1]- CPointY[i],2))
            S = S + L[i]

        tmp = L[0]
        for i in range(4,n):
            tmp = tmp + L[i-3]
            knot[i] = tmp / S
    else:
        tmp = 1 / (n-3)
        tmpS = tmp
        for i in range(4,n):
            knot[i] = tmpS
            tmpS = tmpS + tmp
#     print(knot)
    #节点向量
    #knot = [0,0,0,0,0.25,0.4,0.8,1,1,1,1]

    def de_boor_cal(u,knot,X, Y):
        #判断u在哪个区间
        m = len(knot) #节点个数 且 m = n + k +1
        n = len(X)
        i = 3
        for i in range(3,m-3):
            if u>=knot[i] and u<knot[i+1]:
                break
        # print(i)
        #递归计算
        if knot[i+3] - knot[i] == 0:
            a41 = 0
        else:
            a41 = (u - knot[i]) / (knot[i+3] - knot[i])
        if knot[i-1+3] - knot[i-1] == 0:
            a31 = 0
        else:
            a31 = (u - knot[i-1]) / (knot[i-1+3] - knot[i-1])
        if knot[i-2+3] - knot[i-2] == 0:
            a21 = 0
        else:
            a21 = (u - knot[i-2]) / (knot[i-2+3] - knot[i-2])

        XP40 = X[i]
        YP40 = Y[i]
        XP30 = X[i-1]
        YP30 = Y[i - 1]
        XP20 = X[i-2]
        YP20 = Y[i - 2]
        XP10 = X[i-3]
        YP10 = Y[i - 3]

        XP41 = (1 - a41) * XP30 + a41 * XP40
        YP41 = (1 - a41) * YP30 + a41 * YP40

        XP31 = (1 - a31) * XP20 + a31 * XP30
        YP31 = (1 - a31) * YP20 + a31 * YP30

        XP21 = (1 - a21) * XP10 + a21 * XP20
        YP21 = (1 - a21) * YP10 + a21 * YP20
        if knot[i+3-1] - knot[i] == 0:
            a42 = 0
        else:
            a42 = (u - knot[i]) / (knot[i+3-1] - knot[i])
        if knot[i-1+3-1] - knot[i-1] == 0:
            a32 = 0
        else:
            a32 = (u - knot[i-1]) / (knot[i-1+3-1] - knot[i-1])


        XP42 = (1 - a42) * XP31 + a42 * XP41
        YP42 = (1 - a42) * YP31 + a42 * YP41

        XP32 = (1 - a32) * XP21 + a32 * XP31
        YP32 = (1 - a32) * YP21 + a32 * YP31

        if knot[i+3-2] - knot[i] == 0:
            a43 = 0
        else:
            a43 = (u - knot[i]) / (knot[i+3-2] - knot[i])

        P43 = [(1 - a43) * XP32 + a43 * XP42,(1 - a43) * YP32 + a43 * YP42]
        return P43


    # plt.plot(CPointX,CPointY, '-ob', label="Waypoints")
    # plt.grid(True)
    # plt.legend()
    # plt.axis("equal")

    Bspline = [[],[]]
    for u in np.arange(0.0,1.0,0.0005):
        P = de_boor_cal(u,knot,CPointX,CPointY)
        Bspline[0].append(P[0])
        Bspline[1].append(P[1])
#     ss=0
#     while ss<len(ll_list):
#         u=ll_list[ss]
#         P = de_boor_cal(u,knot,CPointX,CPointY)
#         Bspline[0].append(P[0])
#         Bspline[1].append(P[1])
#         ss+=1

    Bspline[0].append(CPointX[-1])
    Bspline[1].append(CPointY[-1])
    return Bspline




def interpolate(pred_svgpath,target_svgpath,rb=True,savedir=None,save_svg_dir=None,output_dir=None,
                scale=0.75,rotate=0,trans_x=0,trans_y=0,step_num=10,lr=0.1,loss=False,center=False,save_svg=False):


    result_svgs={}

    def mk_dir(path):
        if not os.path.exists(path):
            os.mkdir(path)
        else:
            shutil.rmtree(path)
            os.mkdir(path)
    if save_svg_dir is not None:
        mk_dir(save_svg_dir)
        mk_dir(output_dir)
        print(pred_svgpath, target_svgpath)
        if not os.path.exists(savedir):
            os.mkdir(savedir)
        save_file_path = savedir + os.sep + os.path.split(pred_svgpath)[-1].split(' ')[-1].replace('.svg', '_1.3基础几何' +
                                                                                                   os.path.split(
                                                                                                       target_svgpath)[
                                                                                                       -1].split(' ')[-1])
        save_svg_path = save_svg_dir + os.sep + os.path.split(pred_svgpath)[-1].split(' ')[-1].replace('.svg', '_1.3基础几何' +
                                                                                                       os.path.split(
                                                                                                           target_svgpath)[
                                                                                                           -1].split(' ')[
                                                                                                           -1])

        print(save_file_path)
    svg = SVG.load_svg(pred_svgpath,rb).normalize(Bbox(100)).zoom(0.9).canonicalize().simplify_heuristic()
    svg_pred = SVGTensor.from_data(svg.to_tensor())
    p_pred, inds = svg_pred.sample_points(n=10, mode='demo')

    bbox_pred = find_bbox(p_pred)
    pred_size = max(bbox_pred[1] - bbox_pred[0], bbox_pred[3] - bbox_pred[2])
    size = int(pred_size * (1/scale))
    svg_normsize = Bbox(size)
    svg_rotate = (360-rotate) / 180 * math.pi
    trans_scale = float((0.5*pred_size + 0.5*size) / 60)
    # print(trans_scale,Point(-0.,-0.))
    svg_translate = Point(-trans_x * trans_scale, -trans_y * trans_scale)
    # print(svg_translate)


    tangle = SVG.load_svg(target_svgpath,rb).normalize(svg_normsize).zoom(0.9).canonicalize().simplify_heuristic()
    # print('read tangle',print(tangle.to_tensor().shape))
    svg_target = SVGTensor.from_data(tangle.to_tensor())
    p_target = svg_target.sample_points(n=5)

    # print('read pred target ok')
    pred_c = find_center(p_pred, mode='bbox')
    target_c = find_center(p_target, mode='bbox')

    tangle = tangle.translate(Point(float(pred_c[0] - target_c[0]), float(pred_c[1] - target_c[1]))).rotate(
        svg_rotate,Point(float(pred_c[0]),float(pred_c[1]))).translate(svg_translate).canonicalize().simplify_heuristic()
    svg_target = SVGTensor.from_data(tangle.to_tensor())
    p_target = svg_target.sample_points(n=5)
    # print('center')



    if loss==True:
        l_first, losses = svg_emd_loss(p_pred, p_target, mode='demo')
        print(l_first, losses)
        la = losses[:-1].reshape(-1, 4)
        weight_curve = la.mean(dim=-1)
        weight_curve = (weight_curve - weight_curve.min()) / weight_curve.max() * 5 + 1
        mn = weight_curve.mean()
        # print(weight_curve,len(weight_curve))
        import matplotlib.pyplot as plt
        # plt.plot(list(range(len(weight_curve))),weight_curve)
        pad_ids = np.where(inds == False)
        # print(np.where(inds==False))
        weight_curve = weight_curve.tolist()
        for ids in pad_ids:
            weight_curve.insert(ids[0], mn)

        weight_params = []
        control1 = svg_pred.control1
        control2 = svg_pred.control2
        end_pos = svg_pred.end_pos
        comm = svg_pred.commands
        c1 = []
        c2 = []
        ep = []
        for i in range(len(weight_curve)):
            a = control1[i]
            a.requires_grad_(True)
            b = control2[i]
            b.requires_grad_(True)
            c = end_pos[i]
            c.requires_grad_(True)

            c1.append(a.unsqueeze(0))
            c2.append(b.unsqueeze(0))
            ep.append(c.unsqueeze(0))

            weight_params.append({'params': a, 'lr': weight_curve[i] * lr})
            weight_params.append({'params': b, 'lr': weight_curve[i] * lr})
            weight_params.append({'params': c, 'lr': weight_curve[i] * lr})

        optimizer = optim.Adam(weight_params, lr=lr)

        img_list = []
        # with_viewbox="%d %d %d %d"%(int(-maxsize*0.375),int(-maxsize*0.375),int(maxsize*1.375),int(maxsize*1.375))
        with_viewbox = "0 0 100 100"
        print(with_viewbox)
        best_l = 100
        last_loss = None
        import copy


        for i in range(300):
            optimizer.zero_grad()

            p_pred = sample_points(comm, torch.cat(c1), torch.cat(c2), torch.cat(ep), n=5, mode=None)

            l = svg_emd_loss(p_pred, p_target)
            l.backward()
            optimizer.step()

            if i % 2 == 0:
            # if True:
                #         print(svg_pred.control1.grad) #不一样，就是与距离相关
                # img = svg_pred.draw(with_points=True, do_display=False, return_png=True, with_viewbox=with_viewbox)
                # img_list.append(img)
                # print(l)
                pred_bbox = find_bbox(p_pred)
                bbox_center = Point(float((pred_bbox[0] + pred_bbox[1]) / 2), float((pred_bbox[2] + pred_bbox[3]) / 2))
                svg_pred.save_svg(Bbox(100), save_svg_path.replace('.svg', '_%d.svg' % i), with_viewbox=pred_bbox,
                                  center=bbox_center)
            if best_l - l < 0.001 and i > 10:
                #     if l<(l_first/2):
                break
            if l < best_l:
                best_l = l

        # to_gif(img_list, file_path=save_file_path.replace('.svg', '.gif'))

        svg_list = os.listdir(save_svg_dir)
        svg_nums = len(svg_list)
        print(svg_nums)
        stride = int(svg_nums / 12)
        if stride == 0:
            stride = 1
        for i in range(11):
            if len(glob.glob(save_svg_dir + os.sep + '*_%d.svg' % (stride * (i + 1) * 2))) == 1:
                svgpath = glob.glob(save_svg_dir + os.sep + '*_%d.svg' % (stride * (i + 1) * 2))[0]
                print(svgpath, output_dir)
                shutil.move(svgpath, output_dir)
        # shutil.rmtree(save_svg_dir)
    else:
        def is_clockwise(p):
            start, end = p[:-1], p[1:]
            return torch.stack([start, end], dim=-1).det().sum() > 0

        def make_clockwise(p):
            if not is_clockwise(p):
                return p.flip(dims=[0])
            return p

        def get_length_distribution(p, normalize=True):
            start, end = p[:-1], p[1:]
            length_distr = torch.norm(end - start, dim=-1).cumsum(dim=0)
            length_distr = torch.cat([length_distr.new_zeros(1),
                                      length_distr])

            if normalize:
                length_distr = length_distr / length_distr[-1]

            return length_distr

        # plt.plot(p_pred[:, 0], 30 - p_pred[:, 1])
        # plt.savefig(output_dir + r"\%s.png" % "pred")
        # plt.show()
        n, m = len(p_pred), len(p_target)

        # Make target point lists clockwise
        p_target = make_clockwise(p_target)

        #如果旋转了，重新找左上角：
        if rotate!=0:
            sorts = torch.argsort(p_target[:, 1])
            x_sort = p_target[:, 0][sorts]
            min_x_sort = torch.argmin(x_sort[:3])
            y_sort = sorts[min_x_sort]
            new_p_target = torch.cat([p_target[y_sort:], p_target[:y_sort]])
            p_target = new_p_target

        #x的中心点，如果对称，从上中心点开始顺时针
        ###new find center
        def find_center_index(pred_points, center_x):
            shape_p = pred_points.shape[0]
            shun = True
            sucss = 0
            indx_x = 1
            while sucss == 0:
                if pred_points[:, 0][indx_x] == pred_points[:, 0][-indx_x]:
                    # print('continue find index of x')
                    indx_x += 1
                else:
                    # print('found it !!! index=', indx_x)
                    sucss = 1
            if torch.abs(pred_points[:, 0][indx_x] - center_x) > torch.abs(pred_points[:, 0][-indx_x] - center_x):
                near_pred = pred_points[int(shape_p * 0.75):]
                shun = False
            else:
                near_pred = pred_points[:int(shape_p * 0.25)]
            sorts = torch.argsort(torch.abs(near_pred[:, 0] - center_x))
            # plt.plot(near_pred[:, 0], near_pred[:, 1])
            x_sort = sorts[0]
            if shun == False:
                x_sort = int(shape_p * 0.75) + x_sort
            return x_sort

        if center== True:
            bbox_pred = find_bbox(p_pred)
            xcenter = (bbox_pred[0] + bbox_pred[1]) / 2
            x_sort = find_center_index(p_pred,xcenter)
            new_p_pred = torch.cat([p_pred[x_sort:], p_pred[:x_sort]])
            p_pred = new_p_pred

            bbox_target = find_bbox(p_target)
            xcenter = (bbox_target[0] + bbox_target[1]) / 2
            x_sort = find_center_index(p_target,xcenter)
            new_p_target = torch.cat([p_target[x_sort:], p_target[:x_sort]])
            p_target = new_p_target

        #######################################################################################
        step = 0
        while step < step_num:
            # Compute length distribution
            distr_pred = torch.linspace(0., 1., n).to(p_pred.device)
            distr_target = get_length_distribution(p_target, normalize=True)
            d = torch.cdist(distr_pred.unsqueeze(-1), distr_target.unsqueeze(-1))
            matching = d.argmin(dim=-1)
            p_target_sub = p_target[matching]
            #往target走一步
            l = (p_target_sub - p_pred) / ((step_num + 1) - step) * torch.tensor([[1, 1]], dtype=torch.float).repeat(n,
                                                                                                                     1) + p_pred

            l = np.array(l) #走完一步的轮廓点结果
            import copy
            ll = copy.deepcopy(l)
            ll = spline_smooth(ll) #B样条平滑


            #存png图片
            # import matplotlib.pyplot as plt
            # fig, axe = plt.subplots(1, 1, figsize=(10, 10))
            # #                 axe.plot(l[:,0],30-np.array(l[:,1]),color='black')
            # #                 axe.fill(l[:,0],30-np.array(l[:,1]),color='black')
            # axe.plot(ll[0], 30 - np.array(ll[1]), color='black')
            # axe.fill(ll[0], 30 - np.array(ll[1]), color='black')
            # plt.savefig(output_dir + r"\%d.png" % step)
            # plt.show()

            #存svg
            ll = np.array(ll).T
            p_pred = ll
            pred_bbox = find_bbox(p_pred)
            p_pred[:, 0] = p_pred[:, 0] - (pred_bbox[0] + pred_bbox[1]) / 2
            p_pred[:, 1] = p_pred[:, 1] - (pred_bbox[2] + pred_bbox[3]) / 2
            p_pred = p_pred * (60 / max(pred_bbox[1] - pred_bbox[0], pred_bbox[3] - pred_bbox[2]))
            p_pred = p_pred + 50
            with_viewbox = '0 0 100 100'
            fill_attr = f'fill="black" stroke="black"'
            marker_attr = ''
            path_filling = '1'
            path_svg = '<path {} {} filling="{}" d="{}"></path>'.format(fill_attr, marker_attr, path_filling,
                                                                        "M" + str(p_pred[0, 0]) + " " + str(
                                                                            p_pred[0, 1]) + " " + " ".join(
                                                                            "L" + str(p_pred[i, 0]) + " " + str(
                                                                                p_pred[i, 1]) for i in
                                                                            range(p_pred.shape[0] - 1)) + " Z")
            svg = str((
                f'<svg xmlns="http://www.w3.org/2000/svg" viewBox="{with_viewbox}">'
                f'{""}'
                f'{path_svg}'
                '</svg>')
            )  # height="100px" width="100px"
            if save_svg == True:
                with open(output_dir + os.sep + 'points_%d.svg'%step, 'w') as f:
                    f.write(svg)
            result_svgs[str(step)]=svg

            p_pred = torch.tensor(l)
            print(p_pred.shape, p_target.shape)
            step += 1

        # plt.plot(p_target[:, 0], 30 - p_target[:, 1])
        # plt.savefig(output_dir + r"\%s.png" % "target")
        # plt.show()


        return result_svgs

if __name__=="__main__":
    parser = argparse.ArgumentParser(description="interpolate between two svgs")
    parser.add_argument("--pred_svgpath",default=r'/home/xiemengni/deepsvg-master/dataset/singlepath_svg/3441.svg',help="")
    parser.add_argument("--target_svgpath", default=r'/home/xiemengni/deepsvg-master/dataset/singlepath_svg/2073.svg', help="")
    parser.add_argument("--lr", default=0.1, type=float,help="")
    parser.add_argument("--transfer_x", default=-10, type=float, help="")
    parser.add_argument("--transfer_y", default=-10, type=float, help="")
    parser.add_argument("--rotate", default=90, type=float, help="")
    parser.add_argument("--scale", default=0.7, type=float, help="")
    parser.add_argument("--step_num", default=10, type=int, help="")
    parser.add_argument("--loss", default=False, action='store_true', help="") #点对逼近，为true则迭代
    parser.add_argument("--center", default=False, action='store_true', help="") #中心点对齐，从中心点开始顺时针

    parser.add_argument("--savedir", default='/home/xiemengni/deepsvg-master/dataset/singlepath_svg/interplationOF基础几何与承勋', help="")
    parser.add_argument("--save_svg_dir", default='/home/xiemengni/deepsvg-master/dataset/singlepath_svg/tmp_svg', help="")
    parser.add_argument("--output_dir", default='/home/xiemengni/deepsvg-master/dataset/singlepath_svg/output_3441-3357_svg', help="")
    args = parser.parse_args()

    _=interpolate(args.pred_svgpath, args.target_svgpath, args.savedir, args.save_svg_dir, args.output_dir,scale=args.scale,rotate=args.rotate,
                trans_x=args.transfer_x,trans_y=args.transfer_y,step_num=args.step_num,lr=args.lr,loss=args.loss,center=args.center)






import time


class Timer:
    def __init__(self):
        self.time = time.time()

    def reset(self):
        self.time = time.time()

    def get_elapsed_time(self):
        elapsed_time = time.time() - self.time
        self.reset()
        return elapsed_time



import os
import uuid
import json
import numpy as np
from moviepy.editor import ImageClip, concatenate_videoclips
import shutil

from deepsvg.svglib.svg import SVG
from deepsvg.svglib.geom import Bbox

from ..config import ROOT_DIR


class Frame:
    def __init__(self, index, keyframe=False, svg=None):
        self.index = index
        self.keyframe = keyframe

        if svg is None:
            svg = SVG([], viewbox=Bbox(256))
        self.svg = svg

        self.kivy_bezierpaths = None

    def to_dict(self):
        return {
            "index": self.index,
            "keyframe": self.keyframe
        }

    @staticmethod
    def load_dict(frame):
        f = Frame(frame["index"], frame["keyframe"])
        return f


class DeepSVGProject:
    def __init__(self, name="Title"):
        self.name = name
        self.uid = str(uuid.uuid4())

        self.frames = [Frame(index=0)]

    @property
    def filename(self):
        return os.path.join(ROOT_DIR, f"{self.uid}.json")

    @property
    def base_dir(self):
        base_dir = os.path.join(ROOT_DIR, self.uid)

        if not os.path.exists(base_dir):
            os.makedirs(base_dir)

        return base_dir

    @property
    def cache_dir(self):
        cache_dir = os.path.join(self.base_dir, "cache")

        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        return cache_dir

    def load_project(self, file_path):
        with open(file_path, "r") as f:
            data = json.load(f)

            self.name = data["name"]
            self.uid = data["uid"]

            self.load_frames(data["frames"])

        shutil.rmtree(self.cache_dir)

    def load_frames(self, frames):
        self.frames = [Frame.load_dict(frame) for frame in frames]

        for frame in self.frames:
            frame.svg = SVG.load_svg(os.path.join(self.base_dir, f"{frame.index}.svg"))

    def save_project(self):
        with open(self.filename, "w") as f:
            data = {
                "name": self.name,
                "uid": self.uid,

                "frames": [frame.to_dict() for frame in self.frames]
            }

            json.dump(data, f)

        self.save_frames()

    def save_frames(self):
        for frame in self.frames:
            frame.svg.save_svg(os.path.join(self.base_dir, f"{frame.index}.svg"))

    def export_to_gif(self, frame_duration=0.1, loop_mode=0):
        from .state import LoopMode

        imgs = [frame.svg.copy().normalize().draw(do_display=False, return_png=True) for frame in self.frames]

        if loop_mode == LoopMode.REVERSE:
            imgs = imgs[::-1]
        elif loop_mode == LoopMode.PINGPONG:
            imgs = imgs + imgs[::-1]

        clips = [ImageClip(np.array(img)).set_duration(frame_duration) for img in imgs]

        clip = concatenate_videoclips(clips, method="compose", bg_color=(255, 255, 255))

        file_path = os.path.join(ROOT_DIR, f"{self.uid}.gif")
        clip.write_gif(file_path, fps=24, verbose=False, logger=None)



from deepsvg.config import _Config
from deepsvg import utils
from deepsvg.utils import Stats, TrainVars, Timer
import os
from tensorboardX import SummaryWriter
import torch.nn as nn
import torch
from datetime import datetime
from torch.utils.data import DataLoader
import argparse
import importlib


# Reproducibility
utils.set_seed(42)


def train(cfg: _Config, model_name, experiment_name="", log_dir="./logs", debug=False, resume=False):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    print("Parameters")
    cfg.print_params()

    print("Loading dataset")
    dataset_load_function = importlib.import_module(cfg.dataloader_module).load_dataset
    dataset = dataset_load_function(cfg)
    dataloader = DataLoader(dataset, batch_size=cfg.batch_size, shuffle=True, drop_last=True,
                            num_workers=cfg.loader_num_workers, collate_fn=cfg.collate_fn)
    model = cfg.make_model().to(device)

    if cfg.pretrained_path is not None:
        print(f"Loading pretrained model {cfg.pretrained_path}")
        utils.load_model(cfg.pretrained_path, model)

    stats = Stats(num_steps=cfg.num_steps, num_epochs=cfg.num_epochs, steps_per_epoch=len(dataloader),
                  stats_to_print=cfg.stats_to_print)
    train_vars = TrainVars()
    timer = Timer()

    stats.num_parameters = utils.count_parameters(model)
    print(f"#Parameters: {stats.num_parameters:,}")

    # Summary Writer
    current_time = datetime.now().strftime("%b%d_%H-%M-%S")
    experiment_identifier = f"{model_name}_{experiment_name}_{current_time}"

    summary_writer = SummaryWriter(os.path.join(log_dir, "tensorboard", "debug" if debug else "full", experiment_identifier))
    checkpoint_dir = os.path.join(log_dir, "models", model_name, experiment_name)
    visualization_dir = os.path.join(log_dir, "visualization", model_name, experiment_name)

    cfg.set_train_vars(train_vars, dataloader)

    # Optimizer, lr & warmup schedulers
    optimizers = cfg.make_optimizers(model)
    scheduler_lrs = cfg.make_schedulers(optimizers, epoch_size=len(dataloader))
    scheduler_warmups = cfg.make_warmup_schedulers(optimizers, scheduler_lrs)

    loss_fns = [l.to(device) for l in cfg.make_losses()]

    if resume:
        ckpt_exists = utils.load_ckpt_list(checkpoint_dir, model, None, optimizers, scheduler_lrs, scheduler_warmups, stats, train_vars)

    if resume and ckpt_exists:
        print(f"Resuming model at epoch {stats.epoch+1}")
        stats.num_steps = cfg.num_epochs * len(dataloader)
    else:
        # Run a single forward pass on the single-device model for initialization of some modules
        single_foward_dataloader = DataLoader(dataset, batch_size=cfg.batch_size // cfg.num_gpus, shuffle=True, drop_last=True,
                                      num_workers=cfg.loader_num_workers, collate_fn=cfg.collate_fn)
        data = next(iter(single_foward_dataloader))
        model_args, params_dict = [data[arg].to(device) for arg in cfg.model_args], cfg.get_params(0, 0)
        model(*model_args, params=params_dict)

    model = nn.DataParallel(model)

    epoch_range = utils.infinite_range(stats.epoch) if cfg.num_epochs is None else range(stats.epoch, cfg.num_epochs)
    for epoch in epoch_range:
        print(f"Epoch {epoch+1}")

        for n_iter, data in enumerate(dataloader):
            step = n_iter + epoch * len(dataloader)

            if cfg.num_steps is not None and step > cfg.num_steps:
                return

            model.train()
            model_args = [data[arg].to(device) for arg in cfg.model_args]
            labels = data["label"].to(device) if "label" in data else None
            params_dict, weights_dict = cfg.get_params(step, epoch), cfg.get_weights(step, epoch)

            for i, (loss_fn, optimizer, scheduler_lr, scheduler_warmup, optimizer_start) in enumerate(zip(loss_fns, optimizers, scheduler_lrs, scheduler_warmups, cfg.optimizer_starts), 1):
                optimizer.zero_grad()

                output = model(*model_args, params=params_dict)
                loss_dict = loss_fn(output, labels, weights=weights_dict)

                if step >= optimizer_start:
                    loss_dict["loss"].backward()
                    if cfg.grad_clip is not None:
                        nn.utils.clip_grad_norm_(model.parameters(), cfg.grad_clip)

                    optimizer.step()
                    if scheduler_lr is not None:
                        scheduler_lr.step()
                    if scheduler_warmup is not None:
                        scheduler_warmup.step()

                stats.update_stats_to_print("train", loss_dict.keys())
                stats.update("train", step, epoch, {
                    ("lr" if i == 1 else f"lr_{i}"): optimizer.param_groups[0]['lr'],
                    **loss_dict
                })

            stats.update("train", step, epoch, {
                **weights_dict,
                "time": timer.get_elapsed_time()
            })

            if step % cfg.log_every == 0 and step > 0:
                print(stats.get_summary("train"))
                stats.write_tensorboard(summary_writer, "train")
                summary_writer.flush()

            if step % cfg.val_every == 0 and step > 0:
                model.eval()

                with torch.no_grad():
                    # Visualization
                    output = None
                    cfg.visualize(model, output, train_vars, step, epoch, summary_writer, visualization_dir)

                timer.reset()

            if not debug and step % cfg.ckpt_every == 0 and step > 0:
                utils.save_ckpt_list(checkpoint_dir, model, cfg, optimizers, scheduler_lrs, scheduler_warmups, stats, train_vars)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='DeepSVG Trainer')
    parser.add_argument("--config-module", type=str, required=True)
    parser.add_argument("--log-dir", type=str, default="./logs")
    parser.add_argument("--debug", action="store_true", default=False)
    parser.add_argument("--resume", action="store_true", default=False)

    args = parser.parse_args()

    cfg = importlib.import_module(args.config_module).Config()
    model_name, experiment_name = args.config_module.split(".")[-2:]

    train(cfg, model_name, experiment_name, log_dir=args.log_dir, debug=args.debug, resume=args.resume)



from collections import defaultdict
from collections import deque
import datetime
import torch


class SmoothedValue(object):
    """Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """

    def __init__(self, window_size=20):
        self.deque = deque(maxlen=window_size)
        self.series = []
        self.total = 0.0
        self.count = 0

    def update(self, value):
        self.deque.append(value)
        self.series.append(value)
        self.count += 1
        self.total += value

    @property
    def median(self):
        d = torch.tensor(list(self.deque))
        return d.median().item()

    @property
    def avg(self):
        d = torch.tensor(list(self.deque))
        return d.mean().item()

    @property
    def global_avg(self):
        return self.total / self.count


class Stats:
    def __init__(self, num_steps=None, num_epochs=None, steps_per_epoch=None, stats_to_print=None):
        self.step = self.epoch = 0

        if num_steps is not None:
            self.num_steps = num_steps
        else:
            self.num_steps = num_epochs * steps_per_epoch

        self.stats = {
            "train": defaultdict(SmoothedValue),
        }
        self.stats_to_print = {k: set(v) for k, v in stats_to_print.items()}

    def to_dict(self):
        return self.__dict__

    def load_dict(self, dict):
        for key, val in dict.items():
            setattr(self, key, val)

    def update(self, split, step, epoch, dict):
        self.step = step
        self.epoch = epoch

        for k, v in dict.items():
            if isinstance(v, torch.Tensor):
                v = v.item()
            assert isinstance(v, (float, int))
            self.stats[split][k].update(v)

    def update_stats_to_print(self, split, stats_to_print):
        self.stats_to_print[split].update(stats_to_print)

    def get_summary(self, split):

        if split == "train":
            completion_pct = self.step / self.num_steps * 100
            eta_seconds = self.stats[split].get("time").global_avg * (self.num_steps - self.step)
            eta_string = datetime.timedelta(seconds=int(eta_seconds))

            s = "[{}/{}, {:.1f}%] eta: {}, ".format(self.step, self.num_steps, completion_pct, eta_string)
        else:
            s = f"[Validation, epoch {self.epoch + 1}] "

        return s + ", ".join(f"{stat}: {self.stats[split].get(stat).median:.4f}" for stat in self.stats_to_print[split])

    def write_tensorboard(self, summary_writer, split):
        summary_writer.add_scalar(f"{split}/epoch", self.epoch + 1, self.step)

        for stat in self.stats_to_print[split]:
            summary_writer.add_scalar(f"{split}/{stat}", self.stats[split].get(stat).median, self.step)

    def is_best(self):
        return True



from kivy.uix.textinput import TextInput
from kivy.properties import StringProperty

DEFAULT_PADDING = 6


class AlignedTextInput(TextInput):
    halign = StringProperty('left')
    valign = StringProperty('top')

    def __init__(self, **kwargs):
        self.halign = kwargs.get("halign", "left")
        self.valign = kwargs.get("valign", "top")

        self.bind(on_text=self.on_text)

        super().__init__(**kwargs)

    def on_text(self, instance, value):
        self.redraw()

    def on_size(self, instance, value):
        self.redraw()

    def redraw(self):
        """
        Note: This methods depends on internal variables of its TextInput
        base class (_lines_rects and _refresh_text())
        """

        self._refresh_text(self.text)

        max_size = max(self._lines_rects, key=lambda r: r.size[0]).size
        num_lines = len(self._lines_rects)

        px = [DEFAULT_PADDING, DEFAULT_PADDING]
        py = [DEFAULT_PADDING, DEFAULT_PADDING]

        if self.halign == 'center':
            d = (self.width - max_size[0]) / 2.0 - DEFAULT_PADDING
            px = [d, d]
        elif self.halign == 'right':
            px[0] = self.width - max_size[0] - DEFAULT_PADDING

        if self.valign == 'middle':
            d = (self.height - max_size[1] * num_lines) / 2.0 - DEFAULT_PADDING
            py = [d + 5, d - 5]
        elif self.valign == 'bottom':
            py[0] = self.height - max_size[1] * num_lines - DEFAULT_PADDING

        self.padding_x = px
        self.padding_y = py



import torch
from torch.utils.data import DataLoader
import torch.nn as nn

from configs.deepsvg.hierarchical_ordered import Config

from deepsvg import utils
from deepsvg.svglib.svg import SVG
from deepsvg.difflib.tensor import SVGTensor
from deepsvg.svglib.geom import Bbox
from deepsvg.svgtensor_dataset import load_dataset, SVGFinetuneDataset
from deepsvg.utils.utils import batchify

from .state.project import DeepSVGProject, Frame
from .utils import easein_easeout


device = torch.device("cuda:0"if torch.cuda.is_available() else "cpu")
pretrained_path = "./pretrained/hierarchical_ordered.pth.tar"

cfg = Config()
cfg.model_cfg.dropout = 0.  # for faster convergence
model = cfg.make_model().to(device)
model.eval()


dataset = load_dataset(cfg)


def decode(z):
    commands_y, args_y = model.greedy_sample(z=z)
    tensor_pred = SVGTensor.from_cmd_args(commands_y[0].cpu(), args_y[0].cpu())
    svg_path_sample = SVG.from_tensor(tensor_pred.data, viewbox=Bbox(256))

    return svg_path_sample


def encode_svg(svg):
    data = dataset.get(model_args=[*cfg.model_args, "tensor_grouped"], svg=svg)
    model_args = batchify((data[key] for key in cfg.model_args), device)
    z = model(*model_args, encode_mode=True)
    return z


def interpolate_svg(svg1, svg2, n=10, ease=True):
    z1, z2 = encode_svg(svg1), encode_svg(svg2)

    alphas = torch.linspace(0., 1., n+2)[1:-1]
    if ease:
        alphas = easein_easeout(alphas)

    z_list = [(1 - a) * z1 + a * z2 for a in alphas]
    svgs = [decode(z) for z in z_list]

    return svgs


def finetune_model(project: DeepSVGProject, nb_augmentations=3500):
    keyframe_ids = [i for i, frame in enumerate(project.frames) if frame.keyframe]

    if len(keyframe_ids) < 2:
        return

    svgs = [project.frames[i].svg for i in keyframe_ids]

    utils.load_model(pretrained_path, model)
    print("Finetuning...")
    finetune_dataset = SVGFinetuneDataset(dataset, svgs, frac=1.0, nb_augmentations=nb_augmentations)
    dataloader = DataLoader(finetune_dataset, batch_size=cfg.batch_size, shuffle=True, drop_last=False,
                            num_workers=cfg.loader_num_workers, collate_fn=cfg.collate_fn)

    # Optimizer, lr & warmup schedulers
    optimizers = cfg.make_optimizers(model)
    scheduler_lrs = cfg.make_schedulers(optimizers, epoch_size=len(dataloader))
    scheduler_warmups = cfg.make_warmup_schedulers(optimizers, scheduler_lrs)

    loss_fns = [l.to(device) for l in cfg.make_losses()]

    epoch = 0
    for step, data in enumerate(dataloader):
        model.train()
        model_args = [data[arg].to(device) for arg in cfg.model_args]
        labels = data["label"].to(device) if "label" in data else None
        params_dict, weights_dict = cfg.get_params(step, epoch), cfg.get_weights(step, epoch)

        for i, (loss_fn, optimizer, scheduler_lr, scheduler_warmup, optimizer_start) in enumerate(
                zip(loss_fns, optimizers, scheduler_lrs, scheduler_warmups, cfg.optimizer_starts), 1):
            optimizer.zero_grad()

            output = model(*model_args, params=params_dict)
            loss_dict = loss_fn(output, labels, weights=weights_dict)

            loss_dict["loss"].backward()
            if cfg.grad_clip is not None:
                nn.utils.clip_grad_norm_(model.parameters(), cfg.grad_clip)

            optimizer.step()
            if scheduler_lr is not None:
                scheduler_lr.step()
            if scheduler_warmup is not None:
                scheduler_warmup.step()

            if step % 20 == 0:
                print(f"Step {step}: loss: {loss_dict['loss']}")

    print("Finetuning done.")


def compute_interpolation(project: DeepSVGProject):
    finetune_model(project)

    keyframe_ids = [i for i, frame in enumerate(project.frames) if frame.keyframe]

    if len(keyframe_ids) < 2:
        return

    model.eval()

    for i1, i2 in zip(keyframe_ids[:-1], keyframe_ids[1:]):
        frames_inbetween = i2 - i1 - 1
        if frames_inbetween == 0:
            continue

        svgs = interpolate_svg(project.frames[i1].svg, project.frames[i2].svg, n=frames_inbetween, ease=False)
        for di, svg in enumerate(svgs, 1):
            project.frames[i1 + di] = Frame(i1 + di, keyframe=False, svg=svg)



from __future__ import annotations
from .geom import *
from interpolation.deepsvg.difflib.tensor import SVGTensor
from .util_fns import get_roots
from enum import Enum
import torch
import math
from typing import List, Union
Num = Union[int, float]


class SVGCmdEnum(Enum):
    MOVE_TO = "m"
    LINE_TO = "l"
    CUBIC_BEZIER = "c"
    CLOSE_PATH = "z"
    ELLIPTIC_ARC = "a"
    QUAD_BEZIER = "q"
    LINE_TO_HORIZONTAL = "h"
    LINE_TO_VERTICAL = "v"
    CUBIC_BEZIER_REFL = "s"
    QUAD_BEZIER_REFL = "t"


svgCmdArgTypes = {
    SVGCmdEnum.MOVE_TO.value: [Point],
    SVGCmdEnum.LINE_TO.value: [Point],
    SVGCmdEnum.CUBIC_BEZIER.value: [Point, Point, Point],
    SVGCmdEnum.CLOSE_PATH.value: [],
    SVGCmdEnum.ELLIPTIC_ARC.value: [Radius, Angle, Flag, Flag, Point],
    SVGCmdEnum.QUAD_BEZIER.value: [Point, Point],
    SVGCmdEnum.LINE_TO_HORIZONTAL.value: [XCoord],
    SVGCmdEnum.LINE_TO_VERTICAL.value: [YCoord],
    SVGCmdEnum.CUBIC_BEZIER_REFL.value: [Point, Point],
    SVGCmdEnum.QUAD_BEZIER_REFL.value: [Point],
}


class SVGCommand:
    def __init__(self, command: SVGCmdEnum, args: List[Geom], start_pos: Point, end_pos: Point):
        self.command = command
        self.args = args

        self.start_pos = start_pos
        self.end_pos = end_pos

    def copy(self):
        raise NotImplementedError

    @staticmethod
    def from_str(cmd_str: str, args_str: List[Num], pos=None, initial_pos=None, prev_command: SVGCommand = None):
        if pos is None:
            pos = Point(0.)
        if initial_pos is None:
            initial_pos = Point(0.)

        cmd = SVGCmdEnum(cmd_str.lower())
        # print('command from_str',cmd_str,args_str) M [0.0, 34.64] L [8.91, 21.89, 17.83, 9.14] Z []
        # Implicit MoveTo commands are treated as LineTo
        if cmd is SVGCmdEnum.MOVE_TO and len(args_str) > 2:
            l_cmd_str = SVGCmdEnum.LINE_TO.value
            if cmd_str.isupper():
                l_cmd_str = l_cmd_str.upper()

            l1, pos, initial_pos = SVGCommand.from_str(cmd_str, args_str[:2], pos, initial_pos)
            l2, pos, initial_pos = SVGCommand.from_str(l_cmd_str, args_str[2:], pos, initial_pos)
            return [*l1, *l2], pos, initial_pos

        nb_args = len(args_str)

        if cmd is SVGCmdEnum.CLOSE_PATH:
            assert nb_args == 0, f"Expected no argument for command {cmd_str}: {nb_args} given"
            return [SVGCommandClose(pos, initial_pos)], initial_pos, initial_pos

        expected_nb_args = sum([ArgType.num_args for ArgType in svgCmdArgTypes[cmd.value]])
        assert nb_args % expected_nb_args == 0, f"Expected {expected_nb_args} arguments for command {cmd_str}: {nb_args} given"

        l = []
        i = 0
        for _ in range(nb_args // expected_nb_args):
            args = []
            for ArgType in svgCmdArgTypes[cmd.value]:
                num_args = ArgType.num_args
                arg = ArgType(*args_str[i:i+num_args])

                if cmd_str.islower():
                    arg.translate(pos)
                if isinstance(arg, Coord):
                    arg = arg.to_point(pos)

                args.append(arg)
                i += num_args

            if cmd is SVGCmdEnum.LINE_TO or cmd is SVGCmdEnum.LINE_TO_VERTICAL or cmd is SVGCmdEnum.LINE_TO_HORIZONTAL:
                cmd_parsed = SVGCommandLine(pos, *args)
            elif cmd is SVGCmdEnum.MOVE_TO:
                cmd_parsed = SVGCommandMove(pos, *args)
            elif cmd is SVGCmdEnum.ELLIPTIC_ARC:
                cmd_parsed = SVGCommandArc(pos, *args)
            elif cmd is SVGCmdEnum.CUBIC_BEZIER:
                cmd_parsed = SVGCommandBezier(pos, *args)
            elif cmd is SVGCmdEnum.QUAD_BEZIER:
                Q0 = pos
                Q1 = args[0]
                Q2 = args[1]
                C1 = Point((Q0.x + 2 / 3 * (Q1.x - Q0.x)), (Q0.y + 2 / 3 * (Q1.y - Q0.y)))
                C2 = Point((Q2.x + 2 / 3 * (Q1.x - Q2.x)), (Q2.y + 2 / 3 * (Q1.y - Q2.y)))
                print(C1, C2, args[1])
                cmd_parsed = SVGCommandBezier(pos, C1, C2, args[1])
                # cmd_parsed = SVGCommandBezier(pos, args[0], args[0], args[1])
            elif cmd is SVGCmdEnum.QUAD_BEZIER_REFL or cmd is SVGCmdEnum.CUBIC_BEZIER_REFL:
                if isinstance(prev_command, SVGCommandBezier):
                    control1 = pos * 2 - prev_command.control2
                else:
                    control1 = pos
                control2 = args[0] if cmd is SVGCmdEnum.CUBIC_BEZIER_REFL else control1
                cmd_parsed = SVGCommandBezier(pos, control1, control2, args[-1])

            prev_command = cmd_parsed
            pos = cmd_parsed.end_pos

            if cmd is SVGCmdEnum.MOVE_TO:
                initial_pos = pos

            l.append(cmd_parsed)

        return l, pos, initial_pos

    def __repr__(self):
        cmd = self.command.value.upper()
        return f"{cmd}{self.get_geoms()}"

    def to_str(self):
        cmd = self.command.value.upper()
        return f"{cmd}{' '.join([arg.to_str() for arg in self.args])}"

    def to_tensor(self, PAD_VAL=-1):
        raise NotImplementedError

    @staticmethod
    def from_tensor(vector: torch.Tensor):
        cmd_index, args = int(vector[0]), vector[1:]

        cmd = SVGCmdEnum(SVGTensor.COMMANDS_SIMPLIFIED[cmd_index])
        radius = Radius(*args[:2].tolist())
        x_axis_rotation = Angle(*args[2:3].tolist())
        large_arc_flag = Flag(args[3].item())
        sweep_flag = Flag(args[4].item())
        start_pos = Point(*args[5:7].tolist())
        control1 = Point(*args[7:9].tolist())
        control2 = Point(*args[9:11].tolist())
        end_pos = Point(*args[11:].tolist())

        return SVGCommand.from_args(cmd, radius, x_axis_rotation, large_arc_flag, sweep_flag, start_pos, control1, control2, end_pos)

    @staticmethod
    def from_args(command: SVGCmdEnum, radius: Radius, x_axis_rotation: Angle, large_arc_flag: Flag,
                  sweep_flag: Flag, start_pos: Point, control1: Point, control2: Point, end_pos: Point):
        if command is SVGCmdEnum.MOVE_TO:
            return SVGCommandMove(start_pos, end_pos)
        elif command is SVGCmdEnum.LINE_TO:
            return SVGCommandLine(start_pos, end_pos)
        elif command is SVGCmdEnum.CUBIC_BEZIER:
            return SVGCommandBezier(start_pos, control1, control2, end_pos)
        elif command is SVGCmdEnum.CLOSE_PATH:
            return SVGCommandClose(start_pos, end_pos)
        elif command is SVGCmdEnum.ELLIPTIC_ARC:
            return SVGCommandArc(start_pos, radius, x_axis_rotation, large_arc_flag, sweep_flag, end_pos)

    def draw(self, *args, **kwargs):
        from .svg_path import SVGPath
        return SVGPath([self]).draw(*args, **kwargs)

    def reverse(self):
        raise NotImplementedError

    def is_left_to(self, other: SVGCommand):
        p1, p2 = self.start_pos, other.start_pos

        if p1.y == p2.y:
            return p1.x < p2.x

        return p1.y < p2.y or (np.isclose(p1.norm(), p2.norm()) and p1.x < p2.x)

    def numericalize(self, n=256):
        raise NotImplementedError

    def get_geoms(self):
        return [self.start_pos, self.end_pos]

    def get_points_viz(self, first=False, last=False):
        from .svg_primitive import SVGCircle
        color = "red" if first else "purple" if last else "deepskyblue"  # "#C4C4C4"
        opacity = 0.75 if first or last else 1.0
        return [SVGCircle(self.end_pos, radius=Radius(0.4), color=color, fill=True, stroke_width=".1", opacity=opacity)]

    def get_handles_viz(self):
        return []

    def sample_points(self, n=10, return_array=False):
        return []

    def split(self, n=2):
        raise NotImplementedError

    def length(self):
        raise NotImplementedError

    def bbox(self):
        raise NotImplementedError


class SVGCommandLinear(SVGCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_tensor(self, PAD_VAL=-1):
        cmd_index = SVGTensor.COMMANDS_SIMPLIFIED.index(self.command.value)
        return torch.tensor([cmd_index,
                             *([PAD_VAL] * 5),
                             *self.start_pos.to_tensor(),
                             *([PAD_VAL] * 4),
                             *self.end_pos.to_tensor()])

    def numericalize(self, n=256):
        self.start_pos.numericalize(n)
        self.end_pos.numericalize(n)

    def copy(self):
        return self.__class__(self.start_pos.copy(), self.end_pos.copy())

    def reverse(self):
        return self.__class__(self.end_pos, self.start_pos)

    def split(self, n=2):
        return [self]

    def bbox(self):
        return Bbox(self.start_pos, self.end_pos)


class SVGCommandMove(SVGCommandLinear):
    def __init__(self, start_pos: Point, end_pos: Point=None):
        if end_pos is None:
            start_pos, end_pos = Point(0.), start_pos
        super().__init__(SVGCmdEnum.MOVE_TO, [end_pos], start_pos, end_pos)

    def get_points_viz(self, first=False, last=False):
        from .svg_primitive import SVGLine
        points_viz = super().get_points_viz(first, last)
        points_viz.append(SVGLine(self.start_pos, self.end_pos, color="red", dasharray=0.5))
        return points_viz

    def bbox(self):
        return Bbox(self.end_pos, self.end_pos)


class SVGCommandLine(SVGCommandLinear):
    def __init__(self, start_pos: Point, end_pos: Point):
        super().__init__(SVGCmdEnum.LINE_TO, [end_pos], start_pos, end_pos)

    def sample_points(self, n=10, return_array=False):
        z = np.linspace(0., 1., n)

        if return_array:
            points = (1-z)[:, None] * self.start_pos.pos[None] + z[:, None] * self.end_pos.pos[None]
            return points

        points = [(1 - alpha) * self.start_pos + alpha * self.end_pos for alpha in z]
        return points

    def split(self, n=2):
        points = self.sample_points(n+1)
        return [SVGCommandLine(p1, p2) for p1, p2 in zip(points[:-1], points[1:])]

    def length(self):
        return self.start_pos.dist(self.end_pos)


class SVGCommandClose(SVGCommandLinear):
    def __init__(self, start_pos: Point, end_pos: Point):
        super().__init__(SVGCmdEnum.CLOSE_PATH, [], start_pos, end_pos)

    def get_points_viz(self, first=False, last=False):
        return []


class SVGCommandBezier(SVGCommand):
    def __init__(self, start_pos: Point, control1: Point, control2: Point, end_pos: Point):
        if control2 is None:
            control2 = control1.copy()
        super().__init__(SVGCmdEnum.CUBIC_BEZIER, [control1, control2, end_pos], start_pos, end_pos)

        self.control1 = control1
        self.control2 = control2

    @property
    def p1(self):
        return self.start_pos

    @property
    def p2(self):
        return self.end_pos

    @property
    def q1(self):
        return self.control1

    @property
    def q2(self):
        return self.control2

    def copy(self):
        return SVGCommandBezier(self.start_pos.copy(), self.control1.copy(), self.control2.copy(), self.end_pos.copy())

    def to_tensor(self, PAD_VAL=-1):
        cmd_index = SVGTensor.COMMANDS_SIMPLIFIED.index(SVGCmdEnum.CUBIC_BEZIER.value)
        return torch.tensor([cmd_index,
                             *([PAD_VAL] * 5),
                             *self.start_pos.to_tensor(),
                             *self.control1.to_tensor(),
                             *self.control2.to_tensor(),
                             *self.end_pos.to_tensor()])

    def to_vector(self):
        return np.array([
            self.start_pos.tolist(),
            self.control1.tolist(),
            self.control2.tolist(),
            self.end_pos.tolist()
        ])

    @staticmethod
    def from_vector(vector):
        return SVGCommandBezier(Point(vector[0]), Point(vector[1]), Point(vector[2]), Point(vector[3]))

    def reverse(self):
        return SVGCommandBezier(self.end_pos, self.control2, self.control1, self.start_pos)

    def numericalize(self, n=256):
        self.start_pos.numericalize(n)
        self.control1.numericalize(n)
        self.control2.numericalize(n)
        self.end_pos.numericalize(n)

    def get_geoms(self):
        return [self.start_pos, self.control1, self.control2, self.end_pos]

    def get_handles_viz(self):
        from .svg_primitive import SVGLine, SVGCircle
        anchor_1 = SVGCircle(self.control1, radius=Radius(0.4), color="lime", fill=True, stroke_width=".1")
        anchor_2 = SVGCircle(self.control2, radius=Radius(0.4), color="lime", fill=True, stroke_width=".1")

        handle_1 = SVGLine(self.start_pos, self.control1, color="grey", dasharray=0.5, stroke_width=".1")
        handle_2 = SVGLine(self.end_pos, self.control2, color="grey", dasharray=0.5, stroke_width=".1")
        return [handle_1, handle_2, anchor_1, anchor_2]

    def eval(self, t):
        return (1 - t)**3 * self.start_pos + 3 * (1 - t)**2 * t * self.control1 + 3 * (1 - t) * t**2 * self.control2 + t**3 * self.end_pos

    def derivative(self, t, n=1):
        if n == 1:
            return 3 * (1 - t)**2 * (self.control1 - self.start_pos) + 6 * (1 - t) * t * (self.control2 - self.control1) + 3 * t**2 * (self.end_pos - self.control2)
        elif n == 2:
            return 6 * (1 - t) * (self.control2 - 2 * self.control1 + self.start_pos) + 6 * t * (self.end_pos - 2 * self.control2 + self.control1)

        raise NotImplementedError

    def angle(self, other: SVGCommandBezier):
        t1, t2 = self.derivative(1.), -other.derivative(0.)
        if np.isclose(t1.norm(), 0.) or np.isclose(t2.norm(), 0.):
            return 0.
        angle = np.arccos(np.clip(t1.normalize().dot(t2.normalize()), -1., 1.))
        return np.rad2deg(angle)

    def sample_points(self, n=10, return_array=False):
        b = self.to_vector()

        z = np.linspace(0., 1., n)
        Z = np.stack([np.ones_like(z), z, z**2, z**3], axis=1)
        Q = np.array([[1., 0., 0., 0.],
                      [-3, 3., 0., 0.],
                      [3., -6, 3., 0.],
                      [-1, 3., -3, 1]])

        points = Z @ Q @ b

        if return_array:
            return points

        return [Point(p) for p in points]

    def _split_two(self, z=.5):
        b = self.to_vector()

        Q1 = np.array([[1, 0, 0, 0],
                       [-(z - 1), z, 0, 0],
                       [(z - 1) ** 2, -2 * (z - 1) * z, z ** 2, 0],
                       [-(z - 1) ** 3, 3 * (z - 1) ** 2 * z, -3 * (z - 1) * z ** 2, z ** 3]])
        Q2 = np.array([[-(z - 1) ** 3, 3 * (z - 1) ** 2 * z, -3 * (z - 1) * z ** 2, z ** 3],
                       [0, (z - 1) ** 2, -2 * (z - 1) * z, z ** 2],
                       [0, 0, -(z - 1), z],
                       [0, 0, 0, 1]])

        return SVGCommandBezier.from_vector(Q1 @ b), SVGCommandBezier.from_vector(Q2 @ b)

    def split(self, n=2):
        b_list = []
        b = self

        for i in range(n - 1):
            z = 1. / (n - i)
            b1, b = b._split_two(z)
            b_list.append(b1)
        b_list.append(b)
        return b_list

    def length(self):
        p = self.sample_points(n=100, return_array=True)
        return np.linalg.norm(p[1:] - p[:-1], axis=-1).sum()

    def bbox(self):
        return Bbox.from_points(self.find_extrema())

    def find_roots(self):
        a = 3 * (-self.p1 + 3 * self.q1 - 3 * self.q2 + self.p2)
        b = 6 * (self.p1 - 2 * self.q1 + self.q2)
        c = 3 * (self.q1 - self.p1)

        x_roots, y_roots = get_roots(a.x, b.x, c.x), get_roots(a.y, b.y, c.y)
        roots_cat = [*x_roots, *y_roots]
        roots = [root for root in roots_cat if 0 <= root <= 1]
        return roots

    def find_extrema(self):
        points = [self.start_pos, self.end_pos]
        points.extend([self.eval(root) for root in self.find_roots()])
        return points


class SVGCommandArc(SVGCommand):
    def __init__(self, start_pos: Point, radius: Radius, x_axis_rotation: Angle, large_arc_flag: Flag, sweep_flag: Flag, end_pos: Point):
        super().__init__(SVGCmdEnum.ELLIPTIC_ARC, [radius, x_axis_rotation, large_arc_flag, sweep_flag, end_pos], start_pos, end_pos)

        self.radius = radius
        self.x_axis_rotation = x_axis_rotation
        self.large_arc_flag = large_arc_flag
        self.sweep_flag = sweep_flag

    def copy(self):
        return SVGCommandArc(self.start_pos.copy(), self.radius.copy(), self.x_axis_rotation.copy(), self.large_arc_flag.copy(),
                             self.sweep_flag.copy(), self.end_pos.copy())

    def to_tensor(self, PAD_VAL=-1):
        cmd_index = SVGTensor.COMMANDS_SIMPLIFIED.index(SVGCmdEnum.ELLIPTIC_ARC.value)
        return torch.tensor([cmd_index,
                             *self.radius.to_tensor(),
                             *self.x_axis_rotation.to_tensor(),
                             *self.large_arc_flag.to_tensor(),
                             *self.sweep_flag.to_tensor(),
                             *self.start_pos.to_tensor(),
                             *([PAD_VAL] * 4),
                             *self.end_pos.to_tensor()])

    def _get_center_parametrization(self):
        r = self.radius
        p1, p2 = self.start_pos, self.end_pos

        h, m = 0.5 * (p1 - p2), 0.5 * (p1 + p2)
        p1_trans = h.rotate(-self.x_axis_rotation)

        sign = -1 if self.large_arc_flag.flag == self.sweep_flag.flag else 1
        x2, y2, rx2, ry2 = p1_trans.x**2, p1_trans.y**2, r.x**2, r.y**2
        sqrt = math.sqrt(max((rx2*ry2 - rx2*y2 - ry2*x2) / (rx2*y2 + ry2*x2), 0.))
        c_trans = sign * sqrt * Point(r.x * p1_trans.y / r.y, -r.y * p1_trans.x / r.x)

        c = c_trans.rotate(self.x_axis_rotation) + m

        d, ns = (p1_trans - c_trans) / r, -(p1_trans + c_trans) / r

        theta_1 = Point(1, 0).angle(d, signed=True)

        delta_theta = d.angle(ns, signed=True)
        delta_theta.deg %= 360
        if self.sweep_flag.flag == 0 and delta_theta.deg > 0:
            delta_theta = delta_theta - Angle(360)
        if self.sweep_flag == 1 and delta_theta.deg < 0:
            delta_theta = delta_theta + Angle(360)

        return c, theta_1, delta_theta

    def _get_point(self, c: Point, t: float_type):
        r = self.radius
        return c + Point(r.x * np.cos(t), r.y * np.sin(t)).rotate(self.x_axis_rotation)

    def _get_derivative(self, t: float_type):
        r = self.radius
        return Point(-r.x * np.sin(t), r.y * np.cos(t)).rotate(self.x_axis_rotation)

    def to_beziers(self):
        """ References:
        https://www.w3.org/TR/2018/CR-SVG2-20180807/implnote.html
        https://mortoray.com/2017/02/16/rendering-an-svg-elliptical-arc-as-bezier-curves/
        http://www.spaceroots.org/documents/ellipse/elliptical-arc.pdf """
        beziers = []

        c, theta_1, delta_theta = self._get_center_parametrization()
        nb_curves = max(int(abs(delta_theta.deg) // 45), 1)
        etas = [theta_1 + i * delta_theta / nb_curves for i in range(nb_curves+1)]
        for eta_1, eta_2 in zip(etas[:-1], etas[1:]):
            e1, e2 = eta_1.rad, eta_2.rad
            alpha = np.sin(e2 - e1) * (math.sqrt(4 + 3 * np.tan(0.5 * (e2 - e1))**2) - 1) / 3
            p1, p2 = self._get_point(c, e1), self._get_point(c, e2)
            q1 = p1 + alpha * self._get_derivative(e1)
            q2 = p2 - alpha * self._get_derivative(e2)
            beziers.append(SVGCommandBezier(p1, q1, q2, p2))

        return beziers

    def reverse(self):
        return SVGCommandArc(self.end_pos, self.radius, self.x_axis_rotation, self.large_arc_flag, ~self.sweep_flag, self.start_pos)

    def numericalize(self, n=256):
        raise NotImplementedError

    def get_geoms(self):
        return [self.start_pos, self.radius, self.x_axis_rotation, self.large_arc_flag, self.sweep_flag, self.end_pos]

    def split(self, n=2):
        raise NotImplementedError

    def sample_points(self, n=10, return_array=False):
        raise NotImplementedError



from setuptools import setup
import setuptools

setup(name='svg_func',
      version='1.0.6',
      description='functions in svg for logo',
      author='xmn',
      author_email='2579015983@qq.com',
      url='https://www.python.org/',
      license='MIT',
      keywords='svg',
      project_urls={
            'Documentation': 'https://packaging.python.org/tutorials/distributing-packages/',
            'Funding': 'https://donate.pypi.org',
            'Source': 'https://github.com/pypa/sampleproject/',
            'Tracker': 'https://github.com/pypa/sampleproject/issues',
      },
      package_dir={"": "src"},
      packages=setuptools.find_packages(where="src"),
      install_requires=[],
      python_requires='>=3'
     )



'''
封套扭曲实现
本次1.0版本初步实现5中arch效果
'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
# os.chdir("..")
os.environ['KMP_DUPLICATE_LIB_OK'] = 'TRUE'

from interpolation.deepsvg.svglib.geom import Point
from interpolation.deepsvg.svglib.svg import SVG
from interpolation.deepsvg.svglib.svg_path import SVGPath
from interpolation.deepsvg.svglib.utils import to_gif

from interpolation.deepsvg.difflib.tensor import SVGTensor
from interpolation.deepsvg.difflib.utils import *
from interpolation.deepsvg.difflib.loss import *
from interpolation.deepsvg.svglib.geom import Bbox,Point


import cv2 as cv
import numpy as np
import matplotlib.pyplot as plt
import math
import re
import torch
from xml.dom import expatbuilder

#读取svg里的path
def from_str(svg_str: str):
    svg_path_groups = []

    SVG_list=[]
    svg_dom = expatbuilder.parseString(svg_str, False)
    svg_root = svg_dom.getElementsByTagName('svg')[0]

    viewbox_list = list(map(float, svg_root.getAttribute("viewBox").split(" ")))
    view_box = Bbox(*viewbox_list)

    primitives = {
        "path": SVGPath
    }

    for tag, Primitive in primitives.items():
        for x in svg_dom.getElementsByTagName(tag):
            svg_path_groups.append(Primitive.from_xml(x))
            SVG_list.append(SVG([Primitive.from_xml(x)], view_box))
        return SVG(svg_path_groups, view_box),SVG_list

#读取轮廓点边界框属性
def find_bbox(points):
    return[points[:,0].min(),points[:,0].max(),points[:,1].min(),points[:,1].max(), (points[:,0].min()+points[:,0].max())/2,
          (points[:,1].min()+points[:,1].max())/2,
           points[:,0].max()- points[:,0].min(),
          points[:,1].max()- points[:,1].min()]

#读取svg并取轮廓点，只对单张svg操作
def get_points_color_insvg(svg_path,p_n=10):
    with open(svg_path, "r") as f:
        svg_str = f.read()

    svg,svg_list = from_str(svg_str)
    svg_path_fillcolor = []
    svg_path_indx = []
    svg_path_tensor = []
    for svg_sp in svg_list:
        #每一个path单独处理
        svg_sp = svg_sp.to_path().simplify_arcs()
        svg_path_fillcolor.append(list([svg_sp.svg_path_groups[i].fill for i in range(len(svg_sp.svg_path_groups))]))
        svg_path_indx.append(list(np.where(svg_sp.to_tensor()[:, 0] == 0)[0])+\
                                  [svg_sp.to_tensor().shape[0]])
        svg_path_tensor.append(svg_sp.to_tensor())
        # plot_points(SVGTensor.from_data(svg_sp.to_tensor()).sample_points(n=5), show_color=True)
        # plt.show()
    bh_indx = svg_path_indx
    # print(bh_indx)
    bh_color = svg_path_fillcolor
    svg_tensor = torch.cat(svg_path_tensor,axis=0)
    # print('svg_tensor shape',svg_tensor.shape)
    svg_target = SVGTensor.from_data(svg_tensor)
    p_target = svg_target.sample_points(n=p_n)
    # print('p_target shape:',p_target.shape)

    bbox = find_bbox(p_target)
    return p_target, bh_indx, bh_color, bbox


def trans_points(cont, bbox, n=10, arch_per=0.3, pos=True, fix='top', mode='arch'):
    if mode == 'arch_right':
        cont = torch.flip(cont, [1])
        pos = False
        bbox = find_bbox(cont)

    cont = np.array(cont)
    bbox = np.array(bbox)
    a = [bbox[0], bbox[2]]
    b = [bbox[1], bbox[2]]
    c = [bbox[0], bbox[3]]
    d = [bbox[1], bbox[3]]
    # 角度百分比（与90度比），圆心左右对称有
    arch_a = math.pi / 2 * arch_per
    center_x = (a[0] + b[0]) / 2
    w = (b[0] - a[0])
    h = c[1] - a[1]
    print(bbox)

    if mode=='arch' or mode=='arch_right':
        # xiahu neg
        if pos:
            r = (w / 2) / np.sin(arch_a)  # +0.15#+0.7
            upper_r = r + h
            circle_c = [center_x, c[1] + r * np.cos(arch_a)]
            # print('r', circle_c, r, arch_a)
            # print((cont[0, 0] - center_x) / (w / 2))
            a_cont = ((cont[:, 0] - center_x) / (w / 2)) * arch_a  # left - ,right +  (不均匀)
            r_cont = r + (upper_r - r) * (c[1] - cont[:, 1]) / h  # +
            new_x = circle_c[0] + r_cont * np.sin(a_cont)
            new_y = circle_c[1] - r_cont * np.cos(a_cont)
#             plt.scatter(new_x, new_y)
            # print(bbox, arch_a)
        else:
            r = (w / 2) / np.sin(arch_a)  # +0.15#+0.7
            upper_r = r + h  # *(w/h)
            circle_c = [center_x, a[1] - r * np.cos(arch_a)]
            # print('r', circle_c, r, arch_a)
            a_cont = ((cont[:, 0] - center_x) / (w / 2)) * arch_a  # left - ,right +  (不均匀)
            # print((cont[0, 0] - center_x) / (w / 2))
            r_cont = r + (upper_r - r) * (cont[:, 1] - a[1]) / h  # +
            new_x = circle_c[0] + r_cont * np.sin(a_cont)
            new_y = circle_c[1] + r_cont * np.cos(a_cont)
            # plt.scatter(new_x, new_y)
            # print(bbox, arch_a)
    elif mode=='single_arch':
        if fix == 'top':
            if not pos:
                r = (w / 2) / np.sin(arch_a)  # +0.15#+0.7
                upper_r = r + h
                circle_c = [center_x, c[1] + r * np.cos(arch_a)]
                a_cont = ((cont[:, 0] - center_x) / (w / 2)) * arch_a  # left - ,right +  (不均匀)
                x_arc = circle_c[0] + r * np.sin(a_cont)
                y_arc = circle_c[1] - r * np.abs(np.cos(a_cont))

                y_rate = (cont[:, 1] - a[1]) / h
                new_x = y_rate * cont[:, 0] + (1 - y_rate) * x_arc
                new_y = (y_rate) * a[1] + (1 - y_rate) * y_arc

            else:
                r = (w / 2) / np.sin(arch_a)  # +0.15#+0.7
                upper_r = r + h  # *(w/h)
                circle_c = [center_x, c[1] - r * np.cos(arch_a)]  # c[0]
                a_cont = ((cont[:, 0] - center_x) / (w / 2)) * arch_a  # left - ,right +  (不均匀)
                x_arc = circle_c[0] + r * np.sin(a_cont)
                y_arc = circle_c[1] + r * np.abs(np.cos(a_cont))

                y_rate = (cont[:, 1] - a[1]) / h
                new_x = y_rate * (x_arc - cont[:, 0]) + cont[:, 0]
                new_y = y_rate * (y_arc - a[1]) + a[1]
        else:
            if pos:
                r = (w / 2) / np.sin(arch_a)  # +0.15#+0.7
                upper_r = r + h  # *(w/h)
                circle_c = [center_x, a[1] + r * np.cos(arch_a)]  # c[0]
                a_cont = ((cont[:, 0] - center_x) / (w / 2)) * arch_a  # left - ,right +  (不均匀)
                x_arc = circle_c[0] + r * np.sin(a_cont)
                y_arc = circle_c[1] - r * np.abs(np.cos(a_cont))

                y_rate = (c[1] - cont[:, 1]) / h
                new_x = y_rate * (x_arc - cont[:, 0]) + cont[:, 0]
                new_y = y_rate * (y_arc - c[1]) + c[1]

    transed_points2 = np.array([new_x, new_y])

    if mode == 'arch_right':
        transed_points2 = np.flip(transed_points2, 0)
        w2 = max(transed_points2[0, :]) - min(transed_points2[0, :])
        cx, cy = (max(transed_points2[0, :]) + min(transed_points2[0, :])) / 2, (
                    max(transed_points2[1, :]) + min(transed_points2[1, :])) / 2
        transed_points2[0, :] = (transed_points2[0, :] - cx) * float((bbox[3] - bbox[2])) / w2 + float(
            bbox[3] + bbox[2]) / 2
        transed_points2[1, :] = (transed_points2[1, :] - cy) * float((bbox[3] - bbox[2])) / w2 + float(
            bbox[1] + bbox[0]) / 2
    else:
        w2 = max(transed_points2[0, :]) - min(transed_points2[0, :])
        cx, cy = (max(transed_points2[0, :]) + min(transed_points2[0, :])) / 2, (
                    max(transed_points2[1, :]) + min(transed_points2[1, :])) / 2
        transed_points2[0, :] = (transed_points2[0, :] - cx) * float((bbox[1] - bbox[0])) / w2 + float(
            bbox[1] + bbox[0]) / 2
        transed_points2[1, :] = (transed_points2[1, :] - cy) * float(bbox[1] - bbox[0]) / w2 + float(
            bbox[3] + bbox[2]) / 2
    transed_points2 = np.around(transed_points2,decimals=3)
    # plt.scatter(transed_points2[0,:],transed_points2[1,:])
    return transed_points2

def write_path_insvg(p_pred,color,svg):
    if not color:
        color='black'
    path_svg = '<path d="{}" fill="{}"/>'.format("M" + str(p_pred[0, 0]) + " " + str(
        p_pred[0, 1]) + " " + " ".join(
        "L" + str(p_pred[0, a]) + " " + str(
            p_pred[1, a]) for a in range(p_pred.shape[1] - 1)) + " Z", color)

    svg = svg + str((f'{path_svg}'))
    return svg

def gen_svgstr(p_pred):
    p_str = "M" + str(p_pred[0, 0]) + " " + str(
        p_pred[1, 0]) + " " + " ".join(
        "L" + str(p_pred[0, a]) + " " + str(
            p_pred[1, a]) for a in range(p_pred.shape[1] - 1)) + " Z"
    return p_str

def save_l_svg(transed_points,indx_zi,color_zi, file_path=None,n=10):
    pred_bbox = find_bbox(transed_points.T)
    with_viewbox = str(float(pred_bbox[0]))+' '+str(float(pred_bbox[2]))+' '+str(float(pred_bbox[1]))+' '+str(float(pred_bbox[3]))
    # fill_attr = f'fill="black" stroke="black"'
    fill_attr = ''
    marker_attr = ''
    path_filling = '1'
    svg = str((
        f'<svg xmlns="http://www.w3.org/2000/svg" viewBox="{with_viewbox}">'
        f'{""}'))
    print(color_zi)

    for j in range(len(indx_zi)):
        indx_xiao = indx_zi[j]
        p_str = ''
        print('indx_xiao',indx_xiao)
        p_zi_nums = (indx_xiao[-1]-(len(indx_xiao)-1)*2)*(n-1)
        szi_points = transed_points[:,:p_zi_nums]
        transed_points = transed_points[:,p_zi_nums:]
        color = color_zi[j]
        last_inx = 0
        if len(indx_xiao)>2:
            for i in range(len(indx_xiao) - 2):#path里的笔画
                sinx = indx_xiao[i]
                einx = indx_xiao[i + 1]
                p_num = (einx - sinx - 2) * (n-1)
                print(sinx, einx, p_num,szi_points.shape)
                p_pred = szi_points[:,last_inx:last_inx+p_num]
                p_str += gen_svgstr(p_pred)

                last_inx = last_inx + p_num

                if i == len(indx_xiao) - 3:
                    p_pred = szi_points[:,last_inx:]
                    p_str += gen_svgstr(p_pred)
        else:
            p_pred = szi_points[:,:]
            p_str += gen_svgstr(p_pred)

        if color[0]=="":
            color='black'
        path_svg = '<path d="{}" fill="{}"/>'.format(p_str, color)
        # print(path_svg)
        svg = svg + str((f'{path_svg}'))

    svg = svg + str(('</svg>'))
    # print(svg)
    if file_path is not None:
        with open(file_path, 'w') as f:
            f.write(svg)
    return svg

def main(svgpath,sample_n=10,arch_per=0.2,pos=True,fix='top', mode='arch',file_path = r'C:\Users\25790\Downloads\res.svg'):
    read_points, bh_indx, bh_color, points_bbox = get_points_color_insvg(svgpath, p_n=sample_n)
    transed_points = trans_points(read_points, points_bbox, n=sample_n, arch_per=arch_per, pos=pos, fix=fix, mode=mode)
    svg = save_l_svg(transed_points, bh_indx, bh_color, file_path=file_path)


if __name__ == "__main__":
    main(r"C:\Users\25790\Downloads\封套_业务需求\Artboard 27.svg")




